from __future__ import annotations

from collections import OrderedDict

import numpy as np
import pytest

import dask
from dask.dataframe._compat import PANDAS_GE_220
from dask.dataframe.dask_expr import from_pandas, new_collection
from dask.dataframe.dask_expr._expr import Assign, Blockwise, Filter
from dask.dataframe.dask_expr._reductions import NFirst, NLast
from dask.dataframe.dask_expr._repartition import RepartitionToFewer
from dask.dataframe.dask_expr._shuffle import (
    BaseSetIndexSortValues,
    TaskShuffle,
    divisions_lru,
)
from dask.dataframe.dask_expr.io import FromPandas
from dask.dataframe.dask_expr.tests._util import _backend_library, assert_eq, xfail_gpu
from dask.dataframe.utils import pyarrow_strings_enabled

# Set DataFrame backend for this module
pd = _backend_library()


@pytest.fixture
def pdf():
    return pd.DataFrame({"x": list(range(20)) * 5, "y": range(100)})


@pytest.fixture
def df(pdf):
    return from_pandas(pdf, npartitions=10)


@pytest.mark.parametrize("ignore_index", [True, False])
@pytest.mark.parametrize("npartitions", [3, 6])
def test_disk_shuffle(ignore_index, npartitions, df):
    df2 = df.shuffle(
        "x",
        shuffle_method="disk",
        npartitions=npartitions,
        ignore_index=ignore_index,
    )

    # Check that the output partition count is correct
    assert df2.npartitions == (npartitions or df.npartitions)

    # Check the computed (re-ordered) result
    assert_eq(df, df2, check_index=not ignore_index, check_divisions=False)

    # Check that df was really partitioned by "x".
    # If any values of "x" can be found in multiple
    # partitions, this will fail
    df3 = df2["x"].map_partitions(lambda x: x.drop_duplicates())
    assert sorted(df3.compute().values) == list(range(20))

    # Check `partitions` after shuffle
    a = df2.partitions[1]
    b = df.shuffle(
        "y",
        shuffle_method="disk",
        npartitions=npartitions,
        ignore_index=ignore_index,
    ).partitions[1]
    assert set(a["x"].compute().values.tolist()).issubset(
        b["y"].compute().values.tolist()
    )

    # Check for culling
    assert len(a.optimize().dask) < len(df2.optimize().dask)


@pytest.mark.parametrize("ignore_index", [True, False])
@pytest.mark.parametrize("npartitions", [8, 12])
@pytest.mark.parametrize("max_branch", [32, 6])
def test_task_shuffle(ignore_index, npartitions, max_branch, df):
    df2 = df.shuffle(
        "x",
        shuffle_method="tasks",
        npartitions=npartitions,
        ignore_index=ignore_index,
        max_branch=max_branch,
    )

    # Check that the output partition count is correct
    assert df2.npartitions == (npartitions or df.npartitions)

    # Check the computed (re-ordered) result
    assert_eq(df, df2, check_index=not ignore_index, check_divisions=False)

    # Check that df was really partitioned by "x".
    # If any values of "x" can be found in multiple
    # partitions, this will fail
    df3 = df2["x"].map_partitions(lambda x: x.drop_duplicates())
    assert sorted(df3.compute().values) == list(range(20))

    # Check `partitions` after shuffle
    a = df2.partitions[1]
    b = df.shuffle(
        "y",
        shuffle_method="tasks",
        npartitions=npartitions,
        ignore_index=ignore_index,
    ).partitions[1]
    assert set(a["x"].compute().values.tolist()).issubset(
        b["y"].compute().values.tolist()
    )

    # Check for culling
    assert len(a.optimize().dask) < len(df2.optimize().dask)


@pytest.mark.parametrize("npartitions", [3, 12])
@pytest.mark.parametrize("max_branch", [32, 8])
def test_task_shuffle_index(npartitions, max_branch, pdf):
    pdf = pdf.set_index("x")
    df = from_pandas(pdf, 10)

    df2 = df.shuffle(
        "x",
        shuffle_method="tasks",
        npartitions=npartitions,
        max_branch=max_branch,
    )

    # Check that the output partition count is correct
    assert df2.npartitions == (npartitions or df.npartitions)

    # Check the computed (re-ordered) result
    assert_eq(df, df2, check_divisions=False)

    # Check that df was really partitioned by "x".
    # If any values of "x" can be found in multiple
    # partitions, this will fail
    df3 = df2.index.map_partitions(lambda x: x.drop_duplicates())
    assert sorted(df3.compute().values) == list(range(20))


def test_shuffle_str_column_not_in_dataframe(df):
    with pytest.raises(
        KeyError,
        match="Cannot shuffle on",
    ) as execinfo:
        df.shuffle(on="z")
    assert "z" in str(execinfo.value)


def test_shuffle_mixed_list_column_not_in_dataframe(df):
    with pytest.raises(
        KeyError,
        match="Cannot shuffle on",
    ) as execinfo:
        df.shuffle(["x", "z"])
    assert "z" in str(execinfo.value)
    assert "x" not in str(execinfo.value)


def test_shuffle_list_column_not_in_dataframe(df):
    with pytest.raises(KeyError, match=r"Cannot shuffle on") as excinfo:
        df.shuffle(["zz", "z"])
    assert "z" in str(excinfo.value)
    assert "zz" in str(excinfo.value)


def test_shuffle_column_columns(df):
    df.shuffle(df.columns[-1])


def test_shuffle_column_projection(df):
    df2 = df.shuffle("x")[["x"]].simplify()

    assert "y" not in df2.expr.operands[0].columns


def test_shuffle_reductions(df):
    assert df.shuffle("x").sum().simplify()._name == df.sum()._name


@pytest.mark.xfail(reason="Shuffle can't see the reduction through the Projection")
def test_shuffle_reductions_after_projection(df):
    assert df.shuffle("x").y.sum().simplify()._name == df.y.sum()._name


@pytest.mark.parametrize("ascending", [True, False])
@pytest.mark.parametrize("by", ["a", "b", ["a", "b"]])
@pytest.mark.parametrize("nelem", [10, 500])
def test_sort_values_(nelem, by, ascending):
    np.random.seed(0)
    df = pd.DataFrame()
    df["a"] = np.ascontiguousarray(np.arange(nelem)[::-1])
    df["b"] = np.arange(100, nelem + 100)
    ddf = from_pandas(df, npartitions=10)

    # run on single-threaded scheduler for debugging purposes
    with dask.config.set(scheduler="single-threaded"):
        got = ddf.sort_values(by=by, ascending=ascending)
    expect = df.sort_values(by=by, ascending=ascending)
    assert_eq(got, expect, check_index=False, sort_results=False)


@pytest.mark.parametrize("ascending", [True, False, [False, True], [True, False]])
@pytest.mark.parametrize("by", [["a", "b"], ["b", "a"]])
@pytest.mark.parametrize("nelem", [10, 500])
def test_sort_values_single_partition(nelem, by, ascending):
    np.random.seed(0)
    df = pd.DataFrame()
    df["a"] = np.ascontiguousarray(np.arange(nelem)[::-1])
    df["b"] = np.arange(100, nelem + 100)
    ddf = from_pandas(df, npartitions=1)

    # run on single-threaded scheduler for debugging purposes
    with dask.config.set(scheduler="single-threaded"):
        got = ddf.sort_values(by=by, ascending=ascending)
    expect = df.sort_values(by=by, ascending=ascending)
    assert_eq(got, expect, check_index=False)


@pytest.mark.parametrize("partition_size", [128e6, 128e5])
@pytest.mark.parametrize("upsample", [1.0, 2.0])
def test_set_index(df, pdf, upsample, partition_size):
    assert_eq(
        df.set_index("x", upsample=upsample, partition_size=partition_size),
        pdf.set_index("x"),
    )
    assert_eq(
        df.set_index(df.x, upsample=upsample, partition_size=partition_size),
        pdf.set_index(pdf.x),
    )

    with pytest.raises(NotImplementedError, match="does not yet support"):
        df.set_index(df)


def test_set_index_sorted(pdf):
    pdf = pdf.sort_values(by="y", ignore_index=True)
    pdf["z"] = pdf["x"]
    df = from_pandas(pdf, npartitions=10)
    q = df.set_index("y", sorted=True)
    assert_eq(q, pdf.set_index("y"))

    q = df.set_index("y", sorted=True)["x"]
    assert_eq(q, pdf.set_index("y")["x"])

    with pytest.raises(NotImplementedError, match="not yet support multi-indexes"):
        df.set_index(["y", "z"], sorted=True)

    with pytest.raises(TypeError, match="not supported by set_index"):
        df.set_index([df["y"], df["x"]], sorted=True)

    with pytest.raises(NotImplementedError, match="does not yet support"):
        df.set_index(["y", "z"], sorted=False)


def test_set_index_pre_sorted(pdf):
    pdf = pdf.sort_values(by="y", ignore_index=True)
    pdf["z"] = pdf["x"]
    df = from_pandas(pdf, npartitions=10)
    q = df.set_index("y")
    assert_eq(q, pdf.set_index("y"))
    result = q.optimize(fuse=False).expr
    assert all(isinstance(expr, (Blockwise, FromPandas)) for expr in result.walk())
    q = df.set_index(df.y)
    assert_eq(q, pdf.set_index(pdf.y))
    result = q.optimize(fuse=False).expr
    assert all(isinstance(expr, (Blockwise, FromPandas)) for expr in result.walk())

    q = df.set_index("y")["x"].optimize(fuse=False)
    expected = df[["x", "y"]].set_index("y")["x"].optimize(fuse=False)
    assert q._name == expected._name


@pytest.mark.parametrize("drop", (True, False))
@pytest.mark.parametrize("append", (True, False))
def test_set_index_no_sort(drop, append):
    df = pd.DataFrame({"col1": [2, 4, 1, 3, 5], "col2": [1, 2, 3, 4, 5]})
    ddf = from_pandas(df, npartitions=2)

    assert ddf.npartitions > 1

    # Default is sort=True
    # Index in ddf will be same values, but sorted
    df_result = df.set_index("col1")
    ddf_result = ddf.set_index("col1")
    assert ddf_result.known_divisions
    assert_eq(ddf_result, df_result.sort_index(), sort_results=False)

    # Unknown divisions and index remains unsorted when sort is False
    # and thus equal to pandas set_index, adding extra kwargs also supported by
    # pandas set_index to ensure they're forwarded.
    df_result = df.set_index("col1", drop=drop, append=append)
    ddf_result = ddf.set_index("col1", sort=False, drop=drop, append=append)
    assert not ddf_result.known_divisions
    assert_eq(ddf_result, df_result, sort_results=False)


def test_set_index_repartition(df, pdf):
    result = df.set_index("x", npartitions=2)
    assert result.npartitions == 2
    assert result.optimize(fuse=False).npartitions == 2
    assert_eq(result, pdf.set_index("x"))


def test_set_index_simplify(df, pdf):
    q = df.set_index("x")["y"].optimize(fuse=False)
    expected = df[["x", "y"]].set_index("x")["y"].optimize(fuse=False)
    assert q._name == expected._name

    q = df.set_index(df.x)["y"].optimize(fuse=False)
    expected = df[["y"]].set_index(df.x)["y"].optimize(fuse=False)
    assert q._name == expected._name


def test_set_index_numeric_columns():
    pdf = pd.DataFrame(
        {
            0: list("ABAABBABAA"),
            1: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            2: [1, 2, 3, 2, 1, 3, 2, 4, 2, 3],
        }
    )
    ddf = from_pandas(pdf, 3)
    assert_eq(ddf.set_index(0), pdf.set_index(0))


def test_set_index_without_sort(df, pdf):
    result = df.set_index("y", sort=False)
    assert_eq(result, pdf.set_index("y"))

    result = result.optimize(fuse=False)
    assert all(isinstance(ex, (FromPandas, Blockwise)) for ex in result.walk())

    result = df.set_index(df.y, sort=False)
    assert_eq(result, pdf.set_index(pdf.y))

    with pytest.raises(ValueError, match="Specifying npartitions with sort=False"):
        df.set_index("y", sort=False, npartitions=20)

    with pytest.raises(
        ValueError, match="Specifying npartitions with sort=False or sorted=True"
    ):
        df.set_index("y", sorted=True, npartitions=20)


@pytest.mark.parametrize("shuffle", [None, "tasks"])
def test_sort_values(df, pdf, shuffle):
    assert_eq(df.sort_values("x", shuffle_method=shuffle), pdf.sort_values("x"))
    assert_eq(
        df.sort_values("x", shuffle_method=shuffle, npartitions=2), pdf.sort_values("x")
    )
    pdf.iloc[5, 0] = -10
    df = from_pandas(pdf, npartitions=10)
    assert_eq(
        df.sort_values("x", shuffle_method=shuffle, upsample=2.0), pdf.sort_values("x")
    )

    assert_eq(
        df.sort_values(by=["x", "y"], shuffle_method=shuffle, ascending=[True, True]),
        pdf.sort_values(by=["x", "y"], ascending=[True, True]),
    )
    with pytest.raises(NotImplementedError, match="sorting by named columns"):
        df.sort_values(by=1, shuffle_method=shuffle)

    with pytest.raises(ValueError, match="must be either 'first' or 'last'"):
        df.sort_values(by="x", shuffle_method=shuffle, na_position="bla")


@pytest.mark.parametrize("shuffle", [None, "tasks"])
def test_sort_values_temporary_column_dropped(shuffle):
    pdf = pd.DataFrame(
        {"x": range(10), "y": [1, 2, 3, 4, 5] * 2, "z": ["cat", "dog"] * 5}
    )
    df = from_pandas(pdf, npartitions=2)
    _sorted = df.sort_values(["z"], shuffle_method=shuffle)
    result = _sorted.compute()
    assert "_partitions" not in result.columns


def test_sort_values_optimize(df, pdf):
    q = df.sort_values("x")["y"].optimize(fuse=False)
    expected = df[["x", "y"]].sort_values("x")["y"].optimize(fuse=False)
    assert q._name == expected._name

    q = df.sort_values("x")["x"].optimize(fuse=False)
    expected = df[["x"]].sort_values("x")["x"].optimize(fuse=False)
    assert q._name == expected._name


def test_set_index_single_partition(pdf):
    df = from_pandas(pdf, npartitions=1)
    assert_eq(df.set_index("x"), pdf.set_index("x"))


def test_set_index_list(df, pdf):
    assert_eq(df.set_index(["x"]), pdf.set_index(["x"]))


def test_sort_values_descending(df, pdf):
    assert_eq(
        df.sort_values(by="y", ascending=False),
        pdf.sort_values(by="y", ascending=False),
        sort_results=False,
    )


def test_sort_head_nlargest(df, pdf):
    a = df.sort_values("x", ascending=False).head(10, compute=False)
    b = NFirst(df, 10, _columns=["x"], ascending=False)
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values("x", ascending=True).head(10, compute=False)
    b = NFirst(df, 10, _columns=["x"], ascending=True)
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values("x", ascending=[False]).head(10, compute=False)
    b = NFirst(df, 10, _columns=["x"], ascending=[False])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values("x", ascending=[True]).head(10, compute=False)
    b = NFirst(df, 10, _columns=["x"], ascending=[True])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values(["x"], ascending=[False]).head(10, compute=False)
    b = NFirst(df, 10, _columns=["x"], ascending=[False])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values(["x"], ascending=[True]).head(10, compute=False)
    b = NFirst(df, 10, _columns=["x"], ascending=[True])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values(["x", "y"], ascending=[False, False]).head(10, compute=False)
    b = NFirst(df, 10, _columns=["x", "y"], ascending=[False, False])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values(["x", "y"], ascending=[True, True]).head(10, compute=False).expr
    b = NFirst(df, 10, _columns=["x", "y"], ascending=[True, True])
    assert a.optimize()._name == b.optimize()._name


def test_sort_tail_nsmallest(df, pdf):
    a = df.sort_values("x", ascending=False).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x"], ascending=False)
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values("x", ascending=True).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x"], ascending=True)
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values("x", ascending=[False]).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x"], ascending=[False])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values("x", ascending=[True]).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x"], ascending=[True])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values(["x"], ascending=[False]).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x"], ascending=[False])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values(["x"], ascending=[True]).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x"], ascending=[True])
    assert a.optimize()._name == b.optimize()._name

    with pytest.raises(ValueError, match="Length of ascending"):
        df.sort_values(["x", "y"], ascending=[False]).tail(10, compute=False)

    a = df.sort_values(["x", "y"], ascending=[True, True]).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x", "y"], ascending=[True, True])
    assert a.optimize()._name == b.optimize()._name

    a = df.sort_values(["x", "y"], ascending=[False, False]).tail(10, compute=False)
    b = NLast(df, 10, _columns=["x", "y"], ascending=[False, False])
    assert a.optimize()._name == b.optimize()._name


@pytest.mark.parametrize(
    "ascending",
    [
        pytest.param([True, False], id="[True, False]"),
        pytest.param([False, True], id="[False, True]"),
    ],
)
@pytest.mark.parametrize("npartitions", [1, 3])
def test_sort_values_conflicting_ascending_head_tail(pdf, ascending, npartitions):
    divisions_lru.data = OrderedDict()

    df = from_pandas(pdf, npartitions=npartitions)

    a = df.sort_values(by=["x", "y"], ascending=ascending).head(10, compute=False)
    b = new_collection(NFirst(df, _columns=["x", "y"], n=10, ascending=ascending))
    assert a.expr.optimize()._name == b.expr.optimize()._name
    assert len(divisions_lru) == 0
    assert_eq(
        a.compute(),
        pdf.sort_values(by=["x", "y"], ascending=ascending).head(10),
    )

    a = df.sort_values(by=["x", "y"], ascending=ascending).tail(10, compute=False)
    b = new_collection(NLast(df, _columns=["x", "y"], n=10, ascending=ascending))
    assert a.expr.optimize()._name == b.expr.optimize()._name
    assert len(divisions_lru) == 0
    assert_eq(
        a.compute(),
        pdf.sort_values(by=["x", "y"], ascending=ascending).tail(10),
    )


@xfail_gpu("cudf udf support")
def test_sort_head_nlargest_string(pdf):
    pdf["z"] = "a" + pdf.x.map(str)
    df = from_pandas(pdf, npartitions=5)
    a = df.sort_values("z", ascending=False).head(10, compute=False)
    assert_eq(a, pdf.sort_values("z", ascending=False).head(10))

    a = df.sort_values("z", ascending=True).head(10, compute=False)
    assert_eq(a, pdf.sort_values("z", ascending=True).head(10))

    a = df.sort_values("z", ascending=False).tail(10, compute=False)
    assert_eq(a, pdf.sort_values("z", ascending=False).tail(10))

    a = df.sort_values("z", ascending=True).tail(10, compute=False)
    assert_eq(a, pdf.sort_values("z", ascending=True).tail(10))


def test_set_index_head_nlargest(df, pdf):
    a = df.set_index("x").head(10, compute=False)
    b = new_collection(NFirst(df, 10, _columns="x", ascending=True)).set_index("x")
    assert a.optimize()._name == b.optimize()._name

    a = df.set_index("x").tail(10, compute=False)
    b = new_collection(NLast(df, 10, _columns="x", ascending=True)).set_index("x")
    assert a.optimize()._name == b.optimize()._name

    # These still work, even if we haven't optimized them yet
    df.set_index(df.x).head(3)
    # df.set_index([df.x, df.y]).head(3)


@pytest.mark.skipif(
    not pyarrow_strings_enabled() or not PANDAS_GE_220,
    reason="doesn't work without arrow",
)
@xfail_gpu("cudf udf support")
def test_set_index_head_nlargest_string(pdf):
    pdf["z"] = "a" + pdf.x.map(str)
    df = from_pandas(pdf, npartitions=5)
    print(df.dtypes)

    a = df.set_index("z").head(10, compute=False)
    assert_eq(a, pdf.set_index("z").sort_index().head(10))

    a = df.set_index("z").tail(10, compute=False)
    assert_eq(a, pdf.set_index("z").sort_index().tail(10))


def test_filter_sort(df):
    a = df.sort_values("x")
    a = a[a.y > 40]

    b = df[df.y > 40]
    b = b.sort_values("x")

    assert a.optimize()._name == b.optimize()._name


def test_sort_values_add():
    pdf = pd.DataFrame({"x": [1, 2, 3, 0, 1, 2, 4, 5], "y": 1})
    df = from_pandas(pdf, npartitions=2, sort=False)
    with dask.config.set({"dataframe.shuffle.method": "tasks"}):
        df = df.sort_values("x")
        df["z"] = df.x + df.y
        pdf = pdf.sort_values("x")
        pdf["z"] = pdf.x + pdf.y
        assert_eq(df, pdf, sort_results=False)


@pytest.mark.parametrize("null_value", [None, pd.NaT, pd.NA])
def test_index_nulls(null_value):
    "Setting the index with some non-numeric null raises error"
    df = pd.DataFrame(
        {"numeric": [1, 2, 3, 4], "non_numeric": ["foo", "bar", "foo", "bar"]}
    )
    ddf = from_pandas(df, npartitions=2)
    with pytest.raises(NotImplementedError, match="presence of nulls"):
        ddf.set_index(
            ddf["non_numeric"].map(
                {"foo": "foo", "bar": null_value}, meta=ddf["non_numeric"]._meta
            )
        ).compute()


@pytest.mark.parametrize("freq", ["16h", "-16h"])
def test_set_index_with_dask_dt_index(freq):
    values = {
        "x": [1, 2, 3, 4] * 3,
        "y": [10, 20, 30] * 4,
        "name": ["Alice", "Bob"] * 6,
    }
    date_index = pd.date_range(
        start="2022-02-22", freq=freq, periods=12
    ) - pd.Timedelta(seconds=30)
    df = pd.DataFrame(values, index=date_index)
    ddf = from_pandas(df, npartitions=3, sort=False)
    # specify a different date index entirely
    day_index = ddf.index.dt.floor("D")
    result = ddf.set_index(day_index)
    assert_eq(result, pd.DataFrame(values, index=date_index.floor("D")))


def test_set_index_sort_values_shuffle_options(df, pdf):
    q = df.set_index("x", shuffle_method="tasks", max_branch=10)
    shuffle = list(q.optimize().find_operations(TaskShuffle))[0]
    assert shuffle.options == {"max_branch": 10}
    assert_eq(q, pdf.set_index("x"))

    q = df.sort_values("x", shuffle_method="tasks", max_branch=10)
    sorter = list(q.optimize().find_operations(TaskShuffle))[0]
    assert sorter.options == {"max_branch": 10}
    assert_eq(q, pdf)


def test_set_index_predicate_pushdown(df, pdf):
    pdf = pdf.set_index("x")
    query = df.set_index("x")
    result = query[query.y > 5]
    expected = pdf[pdf.y > 5]
    assert_eq(result, expected)
    expected_query = df[df.y > 5].set_index("x").optimize()
    assert expected_query._name == result.optimize()._name

    result = query[query.index.to_series() > 5]
    assert_eq(result, pdf[pdf.index > 5])

    result = query[(query.index.to_series() > 5) & (query.y > -1)]
    assert_eq(result, pdf[(pdf.index > 5) & (pdf.y > -1)])


def test_set_index_with_explicit_divisions():
    pdf = pd.DataFrame({"x": [4, 1, 2, 5]}, index=[10, 20, 30, 40])

    df = from_pandas(pdf, npartitions=2)

    result = df.set_index("x", divisions=[1, 3, 5])
    assert result.divisions == (1, 3, 5)
    assert_eq(result, pdf.set_index("x"))


def test_set_index_npartitions_changes(pdf):
    df = from_pandas(pdf, npartitions=30)
    result = df.set_index("x", shuffle_method="disk")
    assert result.npartitions == result.optimize().npartitions
    assert_eq(result, pdf.set_index("x"))


def test_set_index_sorted_divisions(df):
    with pytest.raises(ValueError, match="must be the same length"):
        df.set_index("x", divisions=(1, 2, 3), sorted=True)


def test_set_index_sort_values_one_partition(pdf):
    divisions_lru.data = OrderedDict()
    df = from_pandas(pdf, sort=False)
    query = df.sort_values("x").optimize(fuse=False)
    assert query.divisions == (0, 99)
    assert_eq(pdf.sort_values("x"), query, sort_results=False)
    assert len(divisions_lru) == 0

    df = from_pandas(pdf, sort=False)
    query = df.set_index("x").optimize(fuse=False)
    assert query.divisions == (None, None)
    assert_eq(pdf.set_index("x"), query)
    assert len(divisions_lru) == 0

    df = from_pandas(pdf, sort=False, npartitions=2)
    query = df.set_index("x", npartitions=1).optimize(fuse=False)
    assert query.divisions == (None, None)
    assert_eq(pdf.set_index("x"), query)
    assert len(divisions_lru) == 0
    assert len(list(query.expr.find_operations(RepartitionToFewer))) > 0


def test_set_index_triggers_calc_when_accessing_divisions(pdf, df):
    divisions_lru.data = OrderedDict()
    query = df.set_index("x")
    assert len(divisions_lru.data) == 0
    divisions = query.divisions  # noqa: F841
    assert len(divisions_lru.data) == 1


def test_shuffle(df, pdf):
    result = df.shuffle(df.x)
    assert result.npartitions == df.npartitions
    assert_eq(result, pdf)

    result = df.shuffle(df[["x"]])
    assert result.npartitions == df.npartitions
    assert_eq(result, pdf)

    result = df[["y"]].shuffle(df[["x"]])
    assert result.npartitions == df.npartitions
    assert_eq(result, pdf[["y"]])

    with pytest.raises(TypeError, match="index must be aligned"):
        df.shuffle(df.x.repartition(npartitions=2))

    result = df.shuffle(df.x, npartitions=2)
    assert result.npartitions == 2
    assert_eq(result, pdf)


def test_empty_partitions():
    # See https://github.com/dask/dask/issues/2408
    df = pd.DataFrame({"a": list(range(10))})
    df["b"] = df["a"] % 3
    df["c"] = df["b"].astype(str)

    ddf = from_pandas(df, npartitions=3)
    ddf = ddf.set_index("b")
    ddf = ddf.repartition(npartitions=3)
    ddf.get_partition(0).compute()
    assert_eq(ddf, df.set_index("b"))

    ddf = ddf.set_index("c")
    assert_eq(ddf, df.set_index("b").set_index("c"))


def test_shuffle_no_assign(df, pdf):
    result = df.shuffle(df.x)
    q = result.optimize(fuse=False)
    assert len([x for x in q.walk() if isinstance(x, Assign)]) == 0


@pytest.mark.parametrize("meth", ["shuffle", "sort_values"])
def test_shuffle_filter_pushdown(pdf, meth):
    pdf["z"] = 1
    df = from_pandas(pdf, npartitions=10)
    result = getattr(df, meth)("x")
    result = result[result.x > 5.0]
    expected = getattr(df[df.x > 5.0], meth)("x")
    assert result.simplify()._name == expected._name

    result = getattr(df, meth)("x")
    result = result[result.x > 5.0][["x", "y"]]
    expected = df[["x", "y"]]
    expected = getattr(expected[expected.x > 5.0], meth)("x")
    assert result.simplify()._name == expected.simplify()._name

    result = getattr(df, meth)("x")[["x", "y"]]
    result = result[result.x > 5.0]
    expected = df[["x", "y"]]
    expected = getattr(expected[expected.x > 5.0], meth)("x")
    assert result.simplify()._name == expected.simplify()._name


@pytest.mark.parametrize("meth", ["set_index", "sort_values"])
def test_sort_values_avoid_overeager_filter_pushdown(meth):
    pdf1 = pd.DataFrame({"a": [4, 2, 3], "b": [1, 2, 3]})
    df = from_pandas(pdf1, npartitions=2)
    df = getattr(df, meth)("a")
    df = df[df.b > 2] + df.b.sum()
    result = df.simplify()
    assert isinstance(result.expr.left, Filter)
    assert isinstance(result.expr.left.frame, BaseSetIndexSortValues)


def test_set_index_filter_pushdown():
    pdf = pd.DataFrame({"x": [1, 2, 3, 4, 5, 6, 7, 8] * 10, "y": 1, "z": 2})
    df = from_pandas(pdf, npartitions=10)
    result = df.set_index("x")
    result = result[result.y == 1]
    expected = df[df.y == 1].set_index("x")
    assert result.simplify()._name == expected._name

    result = df.set_index("x")
    result = result[result.y == 1][["y"]]
    expected = df[["x", "y"]]
    expected = expected[expected.y == 1].set_index("x")
    assert result.simplify()._name == expected.simplify()._name

    result = df.set_index("x")[["y"]]
    result = result[result.y == 1]
    expected = df[["x", "y"]]
    expected = expected[expected.y == 1].set_index("x")
    assert result.simplify()._name == expected.simplify()._name


def test_shuffle_index_shuffle(df):
    with pytest.raises(TypeError, match="Must shuffle on either "):
        df.shuffle()
    with pytest.raises(TypeError, match="Cannot shuffle on both"):
        df.shuffle("x", on_index=True)


def test_set_index_user_divisions_one_partition(pdf):
    df = from_pandas(pdf, npartitions=1)
    result = df.set_index("x", divisions=[0, 10, 20, 21])
    assert_eq(result, pdf.set_index("x"))


def test_set_index_divisions_npartitions(pdf):
    df = from_pandas(pdf, npartitions=1, sort=False)
    result = df.set_index("x", sort=True, npartitions=2)
    assert result.known_divisions
    assert result.optimize().known_divisions
    assert_eq(result, pdf.set_index("x"))


def test_sort_values_unordered_categorical():
    df = pd.DataFrame()
    df["a"] = np.arange(10)[::-1]
    df["b"] = df["a"].astype("category")
    ddf = from_pandas(df, npartitions=2)
    result = ddf.sort_values(by="b")
    expected = df.sort_values(by="b")
    assert_eq(result, expected, check_index=False)


def test_set_index_before_assign(df, pdf):
    result = df.set_index("x")
    result["z"] = result.y + 1
    expected = pdf.set_index("x")
    expected["z"] = expected.y + 1
    assert_eq(result["z"], expected["z"])


def test_set_index_shuffle_afterwards(pdf):
    ddf = from_pandas(pdf, npartitions=1)
    ddf = ddf.set_index("y", sort=True, divisions=[0, 10, 20, 100], shuffle="tasks")
    result = ddf.reset_index().y.unique()
    expected = pd.Series(pdf.y.unique(), name="y")
    assert_eq(result, expected, check_index=False)



from __future__ import annotations

import importlib
import sys
import types
import warnings
from typing import Literal

if sys.version_info >= (3, 12):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata
from packaging.version import Version

PY_VERSION = Version(".".join(map(str, sys.version_info[:3])))

EMSCRIPTEN = sys.platform == "emscripten"

LINUX = sys.platform == "linux"
MACOS = sys.platform == "darwin"
WINDOWS = sys.platform == "win32"


def entry_points(group=None):
    warnings.warn(
        "`dask._compatibility.entry_points` has been replaced by `importlib_metadata.entry_points` and will be removed "
        "in a future version. Please use `importlib_metadata.entry_points` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return importlib_metadata.entry_points(group=group)


VERSIONS = {
    "numpy": "1.21.0",
    "pandas": "2.0.0",
    "bokeh": "3.1.0",
    "jinja2": "2.10.3",
    "pyarrow": "14.0.1",
    "lz4": "4.3.2",
}

# A mapping from import name to package name (on PyPI) for packages where
# these two names are different.

INSTALL_MAPPING = {
    "sqlalchemy": "SQLAlchemy",
    "tables": "pytables",
}


def get_version(module: types.ModuleType) -> str:
    version = getattr(module, "__version__", None)

    if version is None:
        raise ImportError(f"Can't determine version for {module.__name__}")
    if module.__name__ == "psycopg2":
        # psycopg2 appends " (dt dec pq3 ext lo64)" to it's version
        version = version.split()[0]
    return version


def import_optional_dependency(
    name: str,
    extra: str = "",
    min_version: str | None = None,
    *,
    errors: Literal["raise", "warn", "ignore"] = "raise",
) -> types.ModuleType | None:
    """
    Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice
    message will be raised. If a dependency is present, but too old,
    we raise.

    Parameters
    ----------
    name : str
        The module name.
    extra : str
        Additional text to include in the ImportError message.
    errors : str {'raise', 'warn', 'ignore'}
        What to do when a dependency is not found or its version is too old.

        * raise : Raise an ImportError
        * warn : Only applicable when a module's version is to old.
          Warns that the version is too old and returns None
        * ignore: If the module is not installed, return None, otherwise,
          return the module, even if the version is too old.
          It's expected that users validate the version locally when
          using ``errors="ignore"`` (see. ``io/html.py``)
    min_version : str, default None
        Specify a minimum version that is different from the global pandas
        minimum version required.
    Returns
    -------
    maybe_module : Optional[ModuleType]
        The imported module, when found and the version is correct.
        None is returned when the package is not found and `errors`
        is False, or when the package's version is too old and `errors`
        is ``'warn'`` or ``'ignore'``.
    """
    assert errors in {"warn", "raise", "ignore"}

    package_name = INSTALL_MAPPING.get(name)
    install_name = package_name if package_name is not None else name

    msg = (
        f"Missing optional dependency '{install_name}'. {extra} "
        f"Use pip or conda to install {install_name}."
    )
    try:
        module = importlib.import_module(name)
    except ImportError as err:
        if errors == "raise":
            raise ImportError(msg) from err
        return None

    # Handle submodules: if we have submodule, grab parent module from sys.modules
    parent = name.split(".")[0]
    if parent != name:
        install_name = parent
        module_to_get = sys.modules[install_name]
    else:
        module_to_get = module
    minimum_version = min_version if min_version is not None else VERSIONS.get(parent)
    if minimum_version:
        version = get_version(module_to_get)
        if version and Version(version) < Version(minimum_version):
            msg = (
                f"Dask requires version '{minimum_version}' or newer of '{parent}' "
                f"(version '{version}' currently installed)."
            )
            if errors == "warn":
                warnings.warn(msg, UserWarning)
                return None
            elif errors == "raise":
                raise ImportError(msg)
            else:
                return None

    return module



from __future__ import annotations

import numpy as np
import pytest

from dask.dataframe.dask_expr import from_pandas
from dask.dataframe.dask_expr.io import FromPandas
from dask.dataframe.dask_expr.tests._util import _backend_library, assert_eq

pd = _backend_library()


@pytest.fixture
def pdf():
    pdf = pd.DataFrame({"x": range(20)})
    pdf["y"] = pdf.x * 10.0
    yield pdf


@pytest.fixture
def df(pdf):
    yield from_pandas(pdf, npartitions=4)


def test_iloc(df, pdf):
    assert_eq(df.iloc[:, 1], pdf.iloc[:, 1])
    assert_eq(df.iloc[:, [1]], pdf.iloc[:, [1]])
    assert_eq(df.iloc[:, [0, 1]], pdf.iloc[:, [0, 1]])
    assert_eq(df.iloc[:, []], pdf.iloc[:, []])


def test_iloc_errors(df):
    with pytest.raises(NotImplementedError):
        df.iloc[1]
    with pytest.raises(NotImplementedError):
        df.iloc[1, 1]
    with pytest.raises(ValueError, match="Too many"):
        df.iloc[(1, 2, 3)]


def test_loc_slice(pdf, df):
    pdf.columns = [10, 20]
    df.columns = [10, 20]
    assert_eq(df.loc[:, :15], pdf.loc[:, :15])
    assert_eq(df.loc[:, 15:], pdf.loc[:, 15:])
    assert_eq(df.loc[:, 25:], pdf.loc[:, 25:])  # no columns
    assert_eq(df.loc[:, ::-1], pdf.loc[:, ::-1])


def test_iloc_slice(df, pdf):
    assert_eq(df.iloc[:, :1], pdf.iloc[:, :1])
    assert_eq(df.iloc[:, 1:], pdf.iloc[:, 1:])
    assert_eq(df.iloc[:, 99:], pdf.iloc[:, 99:])  # no columns
    assert_eq(df.iloc[:, ::-1], pdf.iloc[:, ::-1])


@pytest.mark.parametrize("loc", [False, True])
@pytest.mark.parametrize("update", [False, True])
def test_columns_dtype_on_empty_slice(df, pdf, loc, update):
    pdf.columns = [10, 20]
    if update:
        df.columns = [10, 20]
    else:
        df = from_pandas(pdf, npartitions=10)

    assert df.columns.dtype == pdf.columns.dtype
    assert df.compute().columns.dtype == pdf.columns.dtype
    assert_eq(df, pdf)

    if loc:
        df = df.loc[:, []]
        pdf = pdf.loc[:, []]
    else:
        df = df[[]]
        pdf = pdf[[]]

    assert df.columns.dtype == pdf.columns.dtype
    assert df.compute().columns.dtype == pdf.columns.dtype
    assert_eq(df, pdf)


def test_loc(df, pdf):
    assert_eq(df.loc[:, "x"], pdf.loc[:, "x"])
    assert_eq(df.loc[:, ["x"]], pdf.loc[:, ["x"]])
    assert_eq(df.loc[:, []], pdf.loc[:, []])

    assert_eq(df.loc[df.y == 20, "x"], pdf.loc[pdf.y == 20, "x"])
    assert_eq(df.loc[df.y == 20, ["x"]], pdf.loc[pdf.y == 20, ["x"]])
    assert df.loc[3:8].divisions[0] == 3
    assert df.loc[3:8].divisions[-1] == 8

    assert df.loc[5].divisions == (5, 5)

    assert_eq(df.loc[5], pdf.loc[5:5])
    assert_eq(df.loc[3:8], pdf.loc[3:8])
    assert_eq(df.loc[:8], pdf.loc[:8])
    assert_eq(df.loc[3:], pdf.loc[3:])
    assert_eq(df.loc[[5]], pdf.loc[[5]])

    assert_eq(df.x.loc[5], pdf.x.loc[5:5])
    assert_eq(df.x.loc[3:8], pdf.x.loc[3:8])
    assert_eq(df.x.loc[:8], pdf.x.loc[:8])
    assert_eq(df.x.loc[3:], pdf.x.loc[3:])
    assert_eq(df.x.loc[[5]], pdf.x.loc[[5]])
    assert_eq(df.x.loc[[]], pdf.x.loc[[]])
    assert_eq(df.x.loc[np.array([])], pdf.x.loc[np.array([])])

    pytest.raises(KeyError, lambda: df.loc[1000])
    assert_eq(df.loc[1000:], pdf.loc[1000:])
    assert_eq(df.loc[1000:2000], pdf.loc[1000:2000])
    assert_eq(df.loc[:-1000], pdf.loc[:-1000])
    assert_eq(df.loc[-2000:-1000], pdf.loc[-2000:-1000])


def test_loc_non_informative_index():
    df = pd.DataFrame({"x": [1, 2, 3, 4]}, index=[10, 20, 30, 40])
    ddf = from_pandas(df, npartitions=2, sort=True).clear_divisions()
    assert not ddf.known_divisions

    ddf.loc[20:30].compute(scheduler="sync")

    assert_eq(ddf.loc[20:30], df.loc[20:30])

    df = pd.DataFrame({"x": [1, 2, 3, 4]}, index=[10, 20, 20, 40])
    ddf = from_pandas(df, npartitions=2, sort=True)
    assert_eq(ddf.loc[20], df.loc[20:20])


def test_loc_with_series(df, pdf):
    assert_eq(df.loc[df.x % 2 == 0], pdf.loc[pdf.x % 2 == 0])


def test_loc_with_array(df, pdf):
    assert_eq(df.loc[(df.x % 2 == 0).values], pdf.loc[(pdf.x % 2 == 0).values])


def test_loc_with_function(df, pdf):
    assert_eq(df.loc[lambda df: df["x"] > 3, :], pdf.loc[lambda df: df["x"] > 3, :])

    def _col_loc_fun(_df):
        return _df.columns.str.contains("y")

    assert_eq(df.loc[:, _col_loc_fun], pdf.loc[:, _col_loc_fun])


def test_getitem_align():
    df = pd.DataFrame(
        {
            "A": [1, 2, 3, 4, 5, 6, 7, 8, 9],
            "B": [9, 8, 7, 6, 5, 4, 3, 2, 1],
            "C": [True, False, True] * 3,
        },
        columns=list("ABC"),
    )
    ddf = from_pandas(df, 2)
    assert_eq(ddf[ddf.C.repartition([0, 2, 5, 8])], df[df.C])


def test_loc_bool_cindex():
    # https://github.com/dask/dask/issues/11015
    pdf = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    ddf = from_pandas(pdf, npartitions=1)
    indexer = [True, False]
    assert_eq(pdf.loc[:, indexer], ddf.loc[:, indexer])


def test_loc_slicing():
    npartitions = 10
    pdf = pd.DataFrame(
        {
            "A": np.random.randn(npartitions * 10),
        },
        index=pd.date_range("2024-01-01", "2024-12-31", npartitions * 10),
    )
    df = from_pandas(pdf, npartitions=npartitions)
    result = df["2024-03-01":"2024-09-30"]["A"]
    assert_eq(result, pdf["2024-03-01":"2024-09-30"]["A"])


def test_reverse_indexing(df, pdf):
    assert_eq(df.loc[::-1], pdf.loc[::-1])
    result = df.loc[::-1][["x"]]
    expected = pdf.loc[::-1][["x"]]
    assert_eq(result, expected)
    expr = result.expr.optimize(fuse=False)
    assert expr.frame.frame.columns == ["x"] and isinstance(
        expr.frame.frame, FromPandas
    )
    with pytest.raises(
        ValueError, match="Can not use loc on DataFrame without known divisions"
    ):
        df.loc[1::-1].loc[5]
    with pytest.raises(
        ValueError, match="Can not use loc on DataFrame without known divisions"
    ):
        df.loc[:5:-1].loc[5]


def test_indexing_element_index():
    pdf = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    result = from_pandas(pdf, 2).loc[2].index
    pd.testing.assert_index_equal(result.compute(), pdf.loc[[2]].index)

    result = from_pandas(pdf, 2).loc[[2]].index
    pd.testing.assert_index_equal(result.compute(), pdf.loc[[2]].index)



from __future__ import annotations

import pytest

pytest.importorskip("numpy")

import numpy as np

import dask
import dask.array as da
from dask.array import Array
from dask.array.utils import assert_eq
from dask.multiprocessing import _dumps, _loads
from dask.utils import key_split


@pytest.fixture(params=[da.random.RandomState, da.random.default_rng])
def generator_class(request):
    return request.param


def test_generators(generator_class):
    state = generator_class(5)
    x = state.normal(10, 1, size=10, chunks=5)
    assert_eq(x, x)

    state = generator_class(5)
    y = state.normal(10, 1, size=10, chunks=5)
    assert_eq(x, y)


@pytest.mark.parametrize(
    "sd", [None, 42, np.random.PCG64, da.random.Generator(np.random.PCG64)], ids=type
)
def test_default_rng(sd):
    rng = da.random.default_rng(seed=sd)
    assert isinstance(rng, da.random.Generator)


def test_concurrency(generator_class):
    state = generator_class(5)
    x = state.normal(10, 1, size=10, chunks=2)

    state = generator_class(5)
    y = state.normal(10, 1, size=10, chunks=2)
    assert (x.compute(scheduler="processes") == y.compute(scheduler="processes")).all()


def test_doc_randomstate(generator_class):
    assert "mean" in generator_class(5).normal.__doc__


def test_doc_generator():
    x = da.random.default_rng()
    assert "Generator" in str(x)
    assert "Generator" in repr(x)


def test_serializability(generator_class):
    state = generator_class(5)
    x = state.normal(10, 1, size=10, chunks=5)

    y = _loads(_dumps(x))

    assert_eq(x, y)


def test_determinisim_through_dask_values(generator_class):
    samples_1 = generator_class(42).normal(size=1000, chunks=10)
    samples_2 = generator_class(42).normal(size=1000, chunks=10)

    assert set(samples_1.dask) == set(samples_2.dask)
    assert_eq(samples_1, samples_2)


def test_generator_consistent_names(generator_class):
    state1 = generator_class(42)
    state2 = generator_class(42)
    assert sorted(state1.normal(size=(100, 100), chunks=(10, 10)).dask) == sorted(
        state2.normal(size=(100, 100), chunks=(10, 10)).dask
    )
    assert sorted(
        state1.normal(size=100, loc=4.5, scale=5.0, chunks=10).dask
    ) == sorted(state2.normal(size=100, loc=4.5, scale=5.0, chunks=10).dask)


def test_random(generator_class):
    a = generator_class().random((10, 10), chunks=(5, 5))
    assert isinstance(a, Array)
    assert isinstance(a.name, str) and a.name
    assert a.shape == (10, 10)
    assert a.chunks == ((5, 5), (5, 5))

    x = set(np.array(a).flat)

    assert len(x) > 90


def test_parametrized_random_function(generator_class):
    a = generator_class().exponential(1000, (10, 10), chunks=(5, 5))
    assert isinstance(a, Array)
    assert isinstance(a.name, str) and a.name
    assert a.shape == (10, 10)
    assert a.chunks == ((5, 5), (5, 5))

    x = np.array(a)
    assert 10 < x.mean() < 100000

    y = set(x.flat)
    assert len(y) > 90


def test_kwargs(generator_class):
    a = generator_class().normal(loc=10.0, scale=0.1, size=(10, 10), chunks=(5, 5))
    assert isinstance(a, Array)
    x = np.array(a)
    assert 8 < x.mean() < 12


def test_unique_names(generator_class):
    a = generator_class().random((10, 10), chunks=(5, 5))
    b = generator_class().random((10, 10), chunks=(5, 5))

    assert a.name != b.name


def test_docs(generator_class):
    assert "exponential" in generator_class().exponential.__doc__
    assert "exponential" in generator_class().exponential.__name__
    assert "# doctest: +SKIP" in generator_class().normal.__doc__


def test_can_make_really_big_random_array(generator_class):
    generator_class().normal(10, 1, (1000000, 1000000), chunks=(100000, 100000))


def test_random_seed():
    da.random.seed(123)
    x = da.random.normal(size=10, chunks=5)
    y = da.random.normal(size=10, chunks=5)

    da.random.seed(123)
    a = da.random.normal(size=10, chunks=5)
    b = da.random.normal(size=10, chunks=5)

    assert_eq(x, a)
    assert_eq(y, b)


def test_consistent_across_sizes(generator_class):
    x1 = generator_class(123).random(20, chunks=20)
    x2 = generator_class(123).random(100, chunks=20)[:20]
    x3 = generator_class(123).random(200, chunks=20)[:20]
    assert_eq(x1, x2)
    assert_eq(x1, x3)


@pytest.mark.parametrize("sz", [None, 5, (2, 2)], ids=type)
def test_random_all(sz):
    da.random.beta(1, 2, size=sz, chunks=3).compute()
    da.random.binomial(10, 0.5, size=sz, chunks=3).compute()
    da.random.chisquare(1, size=sz, chunks=3).compute()
    da.random.exponential(1, size=sz, chunks=3).compute()
    da.random.f(1, 2, size=sz, chunks=3).compute()
    da.random.gamma(5, 1, size=sz, chunks=3).compute()
    da.random.geometric(1, size=sz, chunks=3).compute()
    da.random.gumbel(1, size=sz, chunks=3).compute()
    da.random.hypergeometric(1, 2, 3, size=sz, chunks=3).compute()
    da.random.laplace(size=sz, chunks=3).compute()
    da.random.logistic(size=sz, chunks=3).compute()
    da.random.lognormal(size=sz, chunks=3).compute()
    da.random.logseries(0.5, size=sz, chunks=3).compute()
    da.random.multinomial(20, [1 / 6.0] * 6, size=sz, chunks=3).compute()
    da.random.negative_binomial(5, 0.5, size=sz, chunks=3).compute()
    da.random.noncentral_chisquare(2, 2, size=sz, chunks=3).compute()

    da.random.noncentral_f(2, 2, 3, size=sz, chunks=3).compute()
    da.random.normal(2, 2, size=sz, chunks=3).compute()
    da.random.pareto(1, size=sz, chunks=3).compute()
    da.random.poisson(size=sz, chunks=3).compute()

    da.random.power(1, size=sz, chunks=3).compute()
    da.random.rayleigh(size=sz, chunks=3).compute()

    da.random.triangular(1, 2, 3, size=sz, chunks=3).compute()
    da.random.uniform(size=sz, chunks=3).compute()
    da.random.vonmises(2, 3, size=sz, chunks=3).compute()
    da.random.wald(1, 2, size=sz, chunks=3).compute()

    da.random.weibull(2, size=sz, chunks=3).compute()
    da.random.zipf(2, size=sz, chunks=3).compute()

    da.random.standard_cauchy(size=sz, chunks=3).compute()
    da.random.standard_exponential(size=sz, chunks=3).compute()
    da.random.standard_gamma(2, size=sz, chunks=3).compute()
    da.random.standard_normal(size=sz, chunks=3).compute()
    da.random.standard_t(2, size=sz, chunks=3).compute()


def test_RandomState_only_funcs():
    da.random.randint(10, size=5, chunks=3).compute()
    with pytest.warns(DeprecationWarning):
        da.random.random_integers(10, size=5, chunks=3).compute()
    da.random.random_sample(10, chunks=3).compute()


@pytest.mark.parametrize("sz", [None, 5, (2, 2)], ids=type)
def test_Generator_only_funcs(sz):
    da.random.default_rng().integers(5, high=15, size=sz, chunks=3).compute()
    da.random.default_rng().multivariate_hypergeometric(
        [16, 8, 4], 6, size=sz, chunks=6
    ).compute()


@pytest.mark.parametrize("sz", [None, 5, (2, 2)], ids=type)
def test_random_all_with_class_methods(generator_class, sz):
    def rnd_test(func, *args, **kwargs):
        a = func(*args, **kwargs)
        assert type(a._meta) == np.ndarray
        assert_eq(a, a)  # Check that _meta and computed arrays match types

    rnd_test(generator_class().beta, 1, 2, size=sz, chunks=3)
    rnd_test(generator_class().binomial, 10, 0.5, size=sz, chunks=3)
    rnd_test(generator_class().chisquare, 1, size=sz, chunks=3)
    rnd_test(generator_class().exponential, 1, size=sz, chunks=3)
    rnd_test(generator_class().f, 1, 2, size=sz, chunks=3)
    rnd_test(generator_class().gamma, 5, 1, size=sz, chunks=3)
    rnd_test(generator_class().geometric, 1, size=sz, chunks=3)
    rnd_test(generator_class().gumbel, 1, size=sz, chunks=3)
    rnd_test(generator_class().hypergeometric, 1, 2, 3, size=sz, chunks=3)
    rnd_test(generator_class().laplace, size=sz, chunks=3)
    rnd_test(generator_class().logistic, size=sz, chunks=3)
    rnd_test(generator_class().lognormal, size=sz, chunks=3)
    rnd_test(generator_class().logseries, 0.5, size=sz, chunks=3)
    rnd_test(generator_class().multinomial, 20, [1 / 6.0] * 6, size=sz, chunks=3)
    rnd_test(generator_class().negative_binomial, 5, 0.5, size=sz, chunks=3)
    rnd_test(generator_class().noncentral_chisquare, 2, 2, size=sz, chunks=3)
    rnd_test(generator_class().noncentral_f, 2, 2, 3, size=sz, chunks=3)
    rnd_test(generator_class().normal, 2, 2, size=sz, chunks=3)
    rnd_test(generator_class().pareto, 1, size=sz, chunks=3)
    rnd_test(generator_class().poisson, size=sz, chunks=3)
    rnd_test(generator_class().power, 1, size=sz, chunks=3)
    rnd_test(generator_class().rayleigh, size=sz, chunks=3)
    rnd_test(generator_class().triangular, 1, 2, 3, size=sz, chunks=3)
    rnd_test(generator_class().uniform, size=sz, chunks=3)
    rnd_test(generator_class().vonmises, 2, 3, size=sz, chunks=3)
    rnd_test(generator_class().wald, 1, 2, size=sz, chunks=3)
    rnd_test(generator_class().weibull, 2, size=sz, chunks=3)
    rnd_test(generator_class().zipf, 2, size=sz, chunks=3)
    rnd_test(generator_class().standard_cauchy, size=sz, chunks=3)
    rnd_test(generator_class().standard_exponential, size=sz, chunks=3)
    rnd_test(generator_class().standard_gamma, 2, size=sz, chunks=3)
    rnd_test(generator_class().standard_normal, size=sz, chunks=3)
    rnd_test(generator_class().standard_t, 2, size=sz, chunks=3)


@pytest.mark.skipif(
    not hasattr(np, "broadcast_to"), reason='requires numpy 1.10 method "broadcast_to"'
)
def test_array_broadcasting(generator_class):
    arr = np.arange(6).reshape((2, 3))
    daones = da.ones((2, 3, 4), chunks=3)
    assert generator_class().poisson(arr, chunks=3).compute().shape == (2, 3)

    for x in (arr, daones):
        y = generator_class().normal(x, 2, chunks=3)
        assert y.shape == x.shape
        assert y.compute().shape == x.shape

    y = generator_class().normal(daones, 2, chunks=3)
    assert set(daones.dask).issubset(set(y.dask))

    assert generator_class().normal(
        np.ones((1, 4)), da.ones((2, 3, 4), chunks=(2, 3, 4)), chunks=(2, 3, 4)
    ).compute().shape == (2, 3, 4)
    assert generator_class().normal(
        scale=np.ones((1, 4)),
        loc=da.ones((2, 3, 4), chunks=(2, 3, 4)),
        size=(2, 2, 3, 4),
        chunks=(2, 2, 3, 4),
    ).compute().shape == (2, 2, 3, 4)

    with pytest.raises(ValueError):
        generator_class().normal(arr, np.ones((3, 1)), size=(2, 3, 4), chunks=3)

    for o in (np.ones(100), da.ones(100, chunks=(50,)), 1):
        a = generator_class().normal(1000 * o, 0.01, chunks=(50,))
        assert 800 < a.mean().compute() < 1200

    # ensure that mis-matched chunks align well
    x = np.arange(10) ** 3
    y = da.from_array(x, chunks=(1,))
    z = generator_class().normal(y, 0.01, chunks=(10,))

    assert 0.8 < z.mean().compute() / x.mean() < 1.2


def test_multinomial(generator_class):
    for size, chunks in [(5, 3), ((5, 4), (2, 3))]:
        x = generator_class().multinomial(20, [1 / 6.0] * 6, size=size, chunks=chunks)
        y = np.random.multinomial(20, [1 / 6.0] * 6, size=size)

        assert x.shape == y.shape == x.compute().shape


def test_choice(generator_class):
    np_generator = {
        da.random.RandomState: np.random.RandomState,
        da.random.default_rng: np.random.default_rng,
    }
    np_dtype = np_generator[generator_class]().choice(1, size=()).dtype
    size = (10, 3)
    chunks = 4
    x = generator_class().choice(3, size=size, chunks=chunks)
    assert x.dtype == np_dtype
    assert x.shape == size
    res = x.compute()
    assert res.dtype == np_dtype
    assert res.shape == size

    py_a = [1, 3, 5, 7, 9]
    np_a = np.array(py_a, dtype="f8")
    da_a = da.from_array(np_a, chunks=2)

    for a in [py_a, np_a, da_a]:
        x = generator_class().choice(a, size=size, chunks=chunks)
        res = x.compute()
        expected_dtype = np.asarray(a).dtype
        assert x.dtype == expected_dtype
        assert res.dtype == expected_dtype
        assert set(np.unique(res)).issubset(np_a)

    np_p = np.array([0, 0.2, 0.2, 0.3, 0.3])
    da_p = da.from_array(np_p, chunks=2)

    for a, p in [(da_a, np_p), (np_a, da_p)]:
        x = generator_class().choice(a, size=size, chunks=chunks, p=p)
        res = x.compute()
        assert x.dtype == np_a.dtype
        assert res.dtype == np_a.dtype
        assert set(np.unique(res)).issubset(np_a[1:])

    np_dtype = np_generator[generator_class]().choice(1, size=(), p=np.array([1])).dtype
    x = generator_class().choice(5, size=size, chunks=chunks, p=np_p)
    res = x.compute()
    assert x.dtype == np_dtype
    assert res.dtype == np_dtype

    errs = [
        (-1, None),  # negative a
        (np_a[:, None], None),  # a must be 1D
        (np_a, np_p[:, None]),  # p must be 1D
        (np_a, np_p[:-2]),  # a and p must match
        (3, np_p),  # a and p must match
        (4, [0.2, 0.2, 0.3]),
    ]  # p must sum to 1

    for a, p in errs:
        with pytest.raises(ValueError):
            generator_class().choice(a, size=size, chunks=chunks, p=p)

    with pytest.raises(NotImplementedError):
        generator_class().choice(da_a, size=size, chunks=chunks, replace=False)

    # axis needs to be row based since a is 1D
    with pytest.raises(ValueError):
        da.random.default_rng().choice(da_a, size=size, chunks=chunks, axis=1)

    # Want to make sure replace=False works for a single-partition output array
    x = generator_class().choice(da_a, size=da_a.shape[0], chunks=-1, replace=False)
    res = x.compute()
    assert len(res) == len(np.unique(res))


def test_create_with_auto_dimensions():
    with dask.config.set({"array.chunk-size": "128MiB"}):
        x = da.random.random((10000, 10000), chunks=(-1, "auto"))
        assert x.chunks == ((10000,), (1677,) * 5 + (1615,))

        y = da.random.random((10000, 10000), chunks="auto")
        assert y.chunks == ((4096, 4096, 1808),) * 2


def test_names():
    name = da.random.normal(0, 1, size=(1000,), chunks=(500,)).name

    assert name.startswith("normal")
    assert len(key_split(name)) < 10


def test_permutation(generator_class):
    x = da.arange(12, chunks=3)
    y = generator_class().permutation(x)

    assert y.shape == x.shape
    assert y.dtype == x.dtype

    y.compute()  # smoke test

    a = generator_class(0)
    b = generator_class(0)
    r1 = a.permutation(x)
    r2 = b.permutation(x)
    assert_eq(r1, r2)

    # Ensure subsequent `permutation` calls return a different result
    # See https://github.com/dask/dask/issues/12029
    r3 = a.permutation(x)
    r4 = b.permutation(x)
    assert_eq(r3, r4)
    with pytest.raises(AssertionError):
        assert_eq(r1, r3)

    x = generator_class().permutation(100)
    assert x.shape == (100,)


def test_auto_chunks(generator_class):
    with dask.config.set({"array.chunk-size": "50 MiB"}):
        x = generator_class().random((10000, 10000))
        assert 4 < x.npartitions < 32


def test_randint_dtype():
    x = da.random.randint(0, 255, size=10, dtype="uint8")
    assert_eq(x, x)
    assert x.dtype == "uint8"
    assert x.compute().dtype == "uint8"


def test_raises_bad_kwarg(generator_class):
    with pytest.raises(Exception) as info:
        generator_class().standard_cauchy(size=(10,), dtype="float64")

    assert "dtype" in str(info.value)


def test_randomstate_kwargs():
    cupy = pytest.importorskip("cupy")

    rs = da.random.RandomState(RandomState=cupy.random.RandomState)
    x = rs.standard_normal((10, 5), dtype=np.float32)
    assert x.dtype == np.float32

    rs = da.random.default_rng(cupy.random.default_rng())
    x = rs.standard_normal((10, 5), dtype=np.float32)
    assert x.dtype == np.float32



from __future__ import annotations

from itertools import product

import pytest

from dask.dataframe._compat import PANDAS_GE_220, PANDAS_GE_300
from dask.dataframe.dask_expr import from_pandas
from dask.dataframe.dask_expr.tests._util import _backend_library, assert_eq

# Set DataFrame backend for this module
pd = _backend_library()


def resample(df, freq, how="mean", **kwargs):
    return getattr(df.resample(freq, **kwargs), how)()


@pytest.fixture
def pdf():
    idx = pd.date_range("2000-01-01", periods=12, freq="min")
    pdf = pd.DataFrame({"foo": range(len(idx))}, index=idx)
    pdf["bar"] = 1
    yield pdf


@pytest.fixture
def df(pdf):
    yield from_pandas(pdf, npartitions=4)


@pytest.mark.parametrize("kwargs", [{}, {"closed": "left"}])
@pytest.mark.parametrize(
    "api",
    [
        "count",
        "prod",
        "mean",
        "sum",
        "min",
        "max",
        "first",
        "last",
        "var",
        "std",
        "size",
        "nunique",
        "median",
        "quantile",
        "ohlc",
        "sem",
    ],
)
def test_resample_apis(df, pdf, api, kwargs):
    result = getattr(df.resample("2min", **kwargs), api)()
    expected = getattr(pdf.resample("2min", **kwargs), api)()
    assert_eq(result, expected)

    # No column output
    if api not in ("size",):
        result = getattr(df.resample("2min"), api)()["foo"]
        expected = getattr(pdf.resample("2min"), api)()["foo"]
        assert_eq(result, expected)

        if api != "ohlc":
            # ohlc actually gives back a DataFrame, so this doesn't work
            q = result.simplify()
            eq = getattr(df["foo"].resample("2min"), api)().simplify()
            assert q._name == eq._name


@pytest.mark.parametrize(
    ["obj", "method", "npartitions", "freq", "closed", "label"],
    list(
        product(
            ["series", "frame"],
            ["count", "mean", "ohlc"],
            [2, 5],
            ["30min", "h", "D", "W"],
            ["right", "left"],
            ["right", "left"],
        )
    ),
)
def test_series_resample(obj, method, npartitions, freq, closed, label):
    if PANDAS_GE_300 and freq == "D" and closed == "right":
        # Temporary xfail until the upstream issue is resolved
        pytest.xfail("https://github.com/pandas-dev/pandas/issues/62200")

    index = pd.date_range("1-1-2000", "2-15-2000", freq="h")
    index = index.union(pd.date_range("4-15-2000", "5-15-2000", freq="h"))
    if obj == "series":
        ps = pd.Series(range(len(index)), index=index)
    elif obj == "frame":
        ps = pd.DataFrame({"a": range(len(index))}, index=index)
    ds = from_pandas(ps, npartitions=npartitions)
    # Series output

    result = resample(ds, freq, how=method, closed=closed, label=label)
    expected = resample(ps, freq, how=method, closed=closed, label=label)

    assert_eq(result, expected, check_dtype=False)

    divisions = result.divisions

    assert expected.index[0] == divisions[0]
    assert expected.index[-1] == divisions[-1]


def test_resample_agg(df, pdf):
    def my_sum(vals, foo=None, *, bar=None):
        return vals.sum()

    result = df.resample("2min").agg(my_sum, "foo", bar="bar")
    expected = pdf.resample("2min").agg(my_sum, "foo", bar="bar")
    assert_eq(result, expected)

    result = df.resample("2min").agg(my_sum)["foo"]
    expected = pdf.resample("2min").agg(my_sum)["foo"]
    assert_eq(result, expected)

    # simplify up disabled for `agg`, function may access other columns
    q = df.resample("2min").agg(my_sum)["foo"].simplify()
    eq = df["foo"].resample("2min").agg(my_sum).simplify()
    assert q._name != eq._name


@pytest.mark.parametrize("method", ["count", "nunique", "size", "sum"])
def test_resample_has_correct_fill_value(method):
    index = pd.date_range("2000-01-01", "2000-02-15", freq="h")
    index = index.union(pd.date_range("4-15-2000", "5-15-2000", freq="h"))
    ps = pd.Series(range(len(index)), index=index)
    ds = from_pandas(ps, npartitions=2)

    assert_eq(
        getattr(ds.resample("30min"), method)(), getattr(ps.resample("30min"), method)()
    )


@pytest.mark.skipif(not PANDAS_GE_220, reason="freq not supported")
def test_resample_divisions_propagation():
    idx = pd.date_range(start="10:00:00.873821", end="10:05:00", freq="0.002s")
    pdf = pd.DataFrame({"data": 1}, index=idx)
    df = from_pandas(pdf, npartitions=10)
    result = df.resample("0.03s").mean()
    result = result.repartition(freq="1D")
    expected = pdf.resample("0.03s").mean()
    assert_eq(result, expected)

    result = df.resample("0.03s").mean().partitions[1]
    expected = pdf.resample("0.03s").mean()[997 : 2 * 997]
    assert_eq(result, expected)



from __future__ import annotations

import functools
import math
import warnings
from collections.abc import Callable

import numpy as np
import pandas as pd
from pandas.core.apply import reconstruct_func, validate_func_kwargs

from dask import is_dask_collection
from dask._task_spec import Task
from dask.core import flatten
from dask.dataframe.core import (
    _concat,
    apply_and_enforce,
    is_dataframe_like,
    is_series_like,
)
from dask.dataframe.dask_expr._collection import (
    FrameBase,
    Index,
    Series,
    new_collection,
)
from dask.dataframe.dask_expr._expr import (
    Assign,
    Blockwise,
    Expr,
    MapPartitions,
    Projection,
    RenameFrame,
    RenameSeries,
    ToFrame,
    _DeepCopy,
    _extract_meta,
    are_co_aligned,
    determine_column_projection,
    no_default,
)
from dask.dataframe.dask_expr._reductions import ApplyConcatApply, Chunk, Reduction
from dask.dataframe.dask_expr._shuffle import RearrangeByColumn
from dask.dataframe.dask_expr._util import (
    PANDAS_GE_300,
    _convert_to_list,
    get_specified_shuffle,
)
from dask.dataframe.dispatch import concat, make_meta, meta_nonempty
from dask.dataframe.groupby import (
    GROUP_KEYS_DEFAULT,
    _agg_finalize,
    _aggregate_docstring,
    _apply_chunk,
    _build_agg_args,
    _cov_agg,
    _cov_chunk,
    _cum_agg_aligned,
    _cum_agg_filled,
    _cumcount_aggregate,
    _determine_levels,
    _groupby_aggregate,
    _groupby_aggregate_spec,
    _groupby_apply_funcs,
    _groupby_get_group,
    _groupby_raise_unaligned,
    _groupby_slice_apply,
    _groupby_slice_shift,
    _groupby_slice_transform,
    _head_aggregate,
    _head_chunk,
    _non_agg_chunk,
    _normalize_spec,
    _nunique_df_chunk,
    _nunique_df_combine,
    _tail_aggregate,
    _tail_chunk,
    _unique_aggregate,
    _value_counts,
    _value_counts_aggregate,
    _var_agg,
    _var_chunk,
)
from dask.dataframe.utils import insert_meta_param_description, is_scalar
from dask.typing import Key
from dask.utils import M, derived_from, is_index_like


def _as_dict(key, value):
    # Utility to convert a single kwarg to a dict.
    # The dict will be empty if the value is None
    return {} if value is None else {key: value}


def _adjust_split_out_for_group_keys(npartitions, by):
    if len(by) == 1:
        return math.ceil(npartitions / 15)
    return math.ceil(npartitions / (10 / (len(by) - 1)))


class Aggregation:
    """User defined groupby-aggregation.

    This class allows users to define their own custom aggregation in terms of
    operations on Pandas dataframes in a map-reduce style. You need to specify
    what operation to do on each chunk of data, how to combine those chunks of
    data together, and then how to finalize the result.

    See :ref:`dataframe.groupby.aggregate` for more.

    Parameters
    ----------
    name : str
        the name of the aggregation. It should be unique, since intermediate
        result will be identified by this name.
    chunk : callable
        a function that will be called with the grouped column of each
        partition, takes a Pandas SeriesGroupBy in input.
        It can either return a single series or a tuple of series.
        The index has to be equal to the groups.
    agg : callable
        a function that will be called to aggregate the results of each chunk.
        Again the argument(s) will be a Pandas SeriesGroupBy. If ``chunk``
        returned a tuple, ``agg`` will be called with all of them as
        individual positional arguments.
    finalize : callable
        an optional finalizer that will be called with the results from the
        aggregation.

    Examples
    --------
    We could implement ``sum`` as follows:

    >>> custom_sum = dd.Aggregation(
    ...     name='custom_sum',
    ...     chunk=lambda s: s.sum(),
    ...     agg=lambda s0: s0.sum()
    ... )  # doctest: +SKIP
    >>> df.groupby('g').agg(custom_sum)  # doctest: +SKIP

    We can implement ``mean`` as follows:

    >>> custom_mean = dd.Aggregation(
    ...     name='custom_mean',
    ...     chunk=lambda s: (s.count(), s.sum()),
    ...     agg=lambda count, sum: (count.sum(), sum.sum()),
    ...     finalize=lambda count, sum: sum / count,
    ... )  # doctest: +SKIP
    >>> df.groupby('g').agg(custom_mean)  # doctest: +SKIP

    Though of course, both of these are built-in and so you don't need to
    implement them yourself.
    """

    def __init__(self, name, chunk, agg, finalize=None):
        self.chunk = chunk
        self.agg = agg
        self.finalize = finalize
        self.__name__ = name


###
### Groupby-aggregation expressions
###


class GroupByBase:
    @functools.cached_property
    def _by_meta(self):
        return [meta_nonempty(x._meta) if isinstance(x, Expr) else x for x in self.by]

    @functools.cached_property
    def _by_columns(self):
        return [x for x in self.by if not isinstance(x, Expr)]

    @property
    def split_by(self):
        return list(
            flatten(
                [[x] if not isinstance(x, Expr) else x.columns for x in self.by],
                container=list,
            )
        )

    @functools.cached_property
    def by(self):
        return self.operands[len(self._parameters) :]

    @functools.cached_property
    def levels(self):
        return _determine_levels(self.by)

    @property
    def shuffle_by_index(self):
        return True


class GroupByChunk(Chunk, GroupByBase):
    _preserves_partitioning_information = True

    @functools.cached_property
    def _args(self) -> list:
        return [self.frame] + self.by

    @functools.cached_property
    def _meta(self):
        args = [
            meta_nonempty(op._meta) if isinstance(op, Expr) else op for op in self._args
        ]
        return make_meta(self.operation(*args, **self._kwargs))


class GroupByApplyConcatApply(ApplyConcatApply, GroupByBase):
    _chunk_cls = GroupByChunk

    @functools.cached_property
    def _meta_chunk(self):
        meta = meta_nonempty(self.frame._meta)
        return self.chunk(meta, *self._by_meta, **self.chunk_kwargs)

    @property
    def _chunk_cls_args(self):
        return self.by

    @property
    def split_out(self):
        if self.operand("split_out") is None:
            return 1
        return super().split_out

    @property
    def _projection_columns(self):
        return self.frame.columns

    def _tune_down(self):
        if self.operand("split_out") is None:
            return self.substitute_parameters(
                {
                    "split_out": functools.partial(
                        _adjust_split_out_for_group_keys, by=self.by
                    )
                }
            )


class SingleAggregation(GroupByApplyConcatApply, GroupByBase):
    """Single groupby aggregation

    This is an abstract class. Sub-classes must implement
    the following methods:

    -   `groupby_chunk`: Applied to each group within
        the `chunk` method of `GroupByApplyConcatApply`
    -   `groupby_aggregate`: Applied to each group within
        the `aggregate` method of `GroupByApplyConcatApply`

    Parameters
    ----------
    frame: Expr
        Dataframe- or series-like expression to group.
    by: str, list or Series
        The key for grouping
    observed:
        Passed through to dataframe backend.
    dropna:
        Whether rows with NA values should be dropped.
    chunk_kwargs:
        Key-word arguments to pass to `groupby_chunk`.
    aggregate_kwargs:
        Key-word arguments to pass to `aggregate_chunk`.
    """

    _parameters = [
        "frame",
        "observed",
        "dropna",
        "chunk_kwargs",
        "aggregate_kwargs",
        "_slice",
        "split_every",
        "split_out",
        "sort",
        "shuffle_method",
    ]
    _defaults = {
        "observed": None,
        "dropna": None,
        "chunk_kwargs": None,
        "aggregate_kwargs": None,
        "_slice": None,
        "split_every": 8,
        "split_out": None,
        "sort": None,
        "shuffle_method": None,
    }

    groupby_chunk: Callable | None = None
    groupby_aggregate: Callable | None = None

    @classmethod
    def chunk(cls, df, *by, **kwargs):
        return _apply_chunk(df, *by, **kwargs)

    @classmethod
    def aggregate(cls, inputs, **kwargs):
        return _groupby_aggregate(_concat(inputs), **kwargs)

    @property
    def chunk_kwargs(self) -> dict:  # type: ignore
        chunk_kwargs = self.operand("chunk_kwargs") or {}
        columns = self._slice
        return {
            "chunk": self.groupby_chunk,
            "columns": columns,
            **_as_dict("observed", self.observed),
            **_as_dict("dropna", self.dropna),
            **chunk_kwargs,
        }

    @property
    def aggregate_kwargs(self) -> dict:  # type: ignore
        aggregate_kwargs = self.operand("aggregate_kwargs") or {}
        groupby_aggregate = self.groupby_aggregate or self.groupby_chunk
        return {
            "aggfunc": groupby_aggregate,
            "levels": self.levels,
            "sort": self.sort,
            **_as_dict("observed", self.observed),
            **_as_dict("dropna", self.dropna),
            **aggregate_kwargs,
        }

    def _simplify_up(self, parent, dependents):
        return groupby_projection(self, parent, dependents)


class GroupbyAggregationBase(GroupByApplyConcatApply, GroupByBase):
    """Base class for groupby aggregation

    This class can be subclassed to perform a general
    groupby aggregation by passing in a `str`, `list` or
    `dict`-based specification using the `arg` operand.

    Parameters
    ----------
    frame: Expr
        Dataframe- or series-like expression to group.
    by: str, list or Series
        The key for grouping
    arg: str, list or dict
        Aggregation spec defining the specific aggregations
        to perform.
    observed:
        Passed through to dataframe backend.
    dropna:
        Whether rows with NA values should be dropped.
    """

    _parameters = [
        "frame",
        "arg",
        "observed",
        "dropna",
        "split_every",
        "split_out",
        "sort",
        "shuffle_method",
        "_slice",
    ]
    _defaults = {
        "observed": None,
        "dropna": None,
        "split_every": 8,
        "split_out": None,
        "sort": None,
        "shuffle_method": None,
        "_slice": None,
    }

    @functools.cached_property
    def spec(self):
        # Converts the `arg` operand into specific
        # chunk, aggregate, and finalizer functions
        if is_dataframe_like(self.frame._meta):
            group_columns = self._by_columns
            if self._slice:
                non_group_columns = self._slice
                if is_scalar(non_group_columns):
                    non_group_columns = [non_group_columns]
            else:
                non_group_columns = [
                    col for col in self.frame.columns if col not in group_columns
                ]
            spec = _normalize_spec(self.arg, non_group_columns)
        elif is_series_like(self.frame._meta):
            if isinstance(self.arg, (list, tuple, dict)):
                spec = _normalize_spec({None: self.arg}, [])
                spec = [
                    (result_column, func, input_column)
                    for ((_, result_column), func, input_column) in spec
                ]

            else:
                spec = _normalize_spec({None: self.arg}, [])
                spec = [
                    (self.frame.columns[0], func, input_column)
                    for (_, func, input_column) in spec
                ]
        else:
            raise ValueError(f"aggregate on unknown object {self.frame._meta}")

        return spec

    @functools.cached_property
    def agg_args(self):
        keys = ["chunk_funcs", "aggregate_funcs", "finalizers"]
        return dict(zip(keys, _build_agg_args(self.spec)))

    def _simplify_down(self):
        if not isinstance(self.arg, dict):
            return

        # Use agg-spec information to add column projection
        required_columns = (
            set(self._by_columns)
            .union(self.arg.keys())
            .intersection(self.frame.columns)
        )
        column_projection = [
            column for column in self.frame.columns if column in required_columns
        ]
        if column_projection != self.frame.columns:
            return type(self)(self.frame[column_projection], *self.operands[1:])


class GroupbyAggregation(GroupbyAggregationBase):
    """Logical groupby aggregation class

    This class lowers itself to concrete implementations for decomposable
    or holistic aggregations.
    """

    @functools.cached_property
    def _meta(self):
        return self._lower()._meta

    @functools.cached_property
    def _is_decomposable(self):
        return not any(s[1] in ("median", np.median) for s in self.spec)

    def _lower(self):
        cls = (
            DecomposableGroupbyAggregation
            if self._is_decomposable
            else HolisticGroupbyAggregation
        )
        return cls(
            self.frame,
            self.arg,
            self.observed,
            self.dropna,
            self.split_every,
            self.split_out,
            self.sort,
            self.shuffle_method,
            self._slice,
            *self.by,
        )


class HolisticGroupbyAggregation(GroupbyAggregationBase):
    """Groupby aggregation for both decomposable and non-decomposable aggregates

    This class always calculates the aggregates by first collecting all the data for
    the groups and then aggregating at once.

    We are always shuffling, so we will never call combine
    """

    @functools.cached_property
    def _meta(self):
        meta = self._meta_chunk
        aggregate = self.aggregate or (lambda x: x)
        aggregate_kwargs = self.aggregate_kwargs
        meta = aggregate([meta], **aggregate_kwargs)
        return make_meta(meta)

    chunk = staticmethod(_non_agg_chunk)

    @property
    def should_shuffle(self):
        return True

    @classmethod
    def aggregate(cls, inputs, **kwargs):
        return _groupby_aggregate_spec(_concat(inputs), **kwargs)

    @property
    def chunk_kwargs(self) -> dict:  # type: ignore
        return {
            "by": self._by_columns,
            "key": [col for col in self.frame.columns if col not in self._by_columns],
            **_as_dict("observed", self.observed),
            **_as_dict("dropna", self.dropna),
        }

    @property
    def aggregate_kwargs(self) -> dict:  # type: ignore
        return {
            "spec": self.arg,
            "levels": _determine_levels(self.by),
            **_as_dict("observed", self.observed),
            **_as_dict("dropna", self.dropna),
        }


class DecomposableGroupbyAggregation(GroupbyAggregationBase):
    """Groupby aggregation for decomposable aggregates

    The results may be calculated via tree or shuffle reduction.
    """

    chunk = staticmethod(_groupby_apply_funcs)

    @classmethod
    def combine(cls, inputs, **kwargs):
        return _groupby_apply_funcs(_concat(inputs), **kwargs)

    @classmethod
    def aggregate(cls, inputs, **kwargs):
        return _agg_finalize(_concat(inputs), **kwargs)

    @property
    def chunk_kwargs(self) -> dict:  # type: ignore
        return {
            "funcs": self.agg_args["chunk_funcs"],
            "sort": self.sort,
            **_as_dict("observed", self.observed),
            **_as_dict("dropna", self.dropna),
        }

    @property
    def combine_kwargs(self) -> dict:  # type: ignore
        return {
            "funcs": self.agg_args["aggregate_funcs"],
            "level": self.levels,
            "sort": self.sort,
            **_as_dict("observed", self.observed),
            **_as_dict("dropna", self.dropna),
        }

    @property
    def aggregate_kwargs(self) -> dict:  # type: ignore
        return {
            "aggregate_funcs": self.agg_args["aggregate_funcs"],
            "arg": self.arg,
            "columns": self._slice,
            "finalize_funcs": self.agg_args["finalizers"],
            "is_series": self.frame._meta.ndim == 1,
            "level": self.levels,
            "sort": self.sort,
            **_as_dict("observed", self.observed),
            **_as_dict("dropna", self.dropna),
        }


class Sum(SingleAggregation):
    groupby_chunk = M.sum


class Prod(SingleAggregation):
    groupby_chunk = M.prod


class Min(SingleAggregation):
    groupby_chunk = M.min


class Max(SingleAggregation):
    groupby_chunk = M.max


class First(SingleAggregation):
    groupby_chunk = M.first


class Last(SingleAggregation):
    groupby_chunk = M.last


class Count(SingleAggregation):
    groupby_chunk = M.count
    groupby_aggregate = M.sum


class Size(SingleAggregation):
    groupby_chunk = M.size
    groupby_aggregate = M.sum

    def _simplify_down(self):
        if (
            self._slice is not None
            and not isinstance(self._slice, list)
            or self.frame.ndim == 1
        ):
            # Scalar slices influence the result and are allowed, i.e., the name of
            # the series is different
            return

        # We can remove every column since pandas reduces to a Series anyway
        by_columns = self._by_columns
        by_columns = [c for c in by_columns if c in self.frame.columns]
        if set(by_columns) == set(self.frame.columns):
            return

        slice_idx = self._parameters.index("_slice")
        ops = [op if i != slice_idx else None for i, op in enumerate(self.operands)]
        return type(self)(self.frame[by_columns], *ops[1:])


class IdxMin(SingleAggregation):
    groupby_chunk = M.idxmin
    groupby_aggregate = M.first


class IdxMax(IdxMin):
    groupby_chunk = M.idxmax
    groupby_aggregate = M.first


class ValueCounts(SingleAggregation):
    groupby_chunk = staticmethod(_value_counts)
    groupby_aggregate = staticmethod(_value_counts_aggregate)


class Unique(SingleAggregation):
    groupby_chunk = M.unique
    groupby_aggregate = staticmethod(_unique_aggregate)

    @functools.cached_property
    def aggregate_kwargs(self) -> dict:  # type: ignore
        kwargs = super().aggregate_kwargs
        meta = self.frame._meta
        if meta.ndim == 1:
            name = meta.name
        else:
            name = meta[self._slice].name
        return {**kwargs, "name": name}


class Cov(SingleAggregation):
    chunk = staticmethod(_cov_chunk)
    std = False

    @classmethod
    def combine(cls, g, levels):
        return _concat(g)

    @classmethod
    def aggregate(cls, inputs, **kwargs):
        return _cov_agg(_concat(inputs), **kwargs)

    @property
    def chunk_kwargs(self) -> dict:  # type: ignore
        return self.operand("chunk_kwargs")

    @property
    def aggregate_kwargs(self) -> dict:  # type: ignore
        kwargs = self.operand("aggregate_kwargs").copy()
        kwargs["sort"] = self.sort
        kwargs["std"] = self.std
        kwargs["levels"] = self.levels
        return kwargs

    @property
    def combine_kwargs(self) -> dict:  # type: ignore
        return {"levels": self.levels}


class Corr(Cov):
    std = True


class GroupByReduction(Reduction, GroupByBase):
    _chunk_cls = GroupByChunk

    def _tune_down(self):
        if self.operand("split_out") is None:
            return self.substitute_parameters(
                {
                    "split_out": functools.partial(
                        _adjust_split_out_for_group_keys, by=self.by
                    )
                }
            )

    @property
    def split_out(self):
        if self.operand("split_out") is None:
            return 1
        return super().split_out

    @property
    def _chunk_cls_args(self):
        return self.by

    @functools.cached_property
    def _meta_chunk(self):
        meta = meta_nonempty(self.frame._meta)
        return self.chunk(meta, *self._by_meta, **self.chunk_kwargs)

    def _divisions(self):
        if self.sort:
            return (None, None)
        split_out = self.split_out
        if split_out is True:
            split_out = self.frame.npartitions
        return (None,) * (split_out + 1)

    def _simplify_up(self, parent, dependents):
        return groupby_projection(self, parent, dependents)

    @functools.cached_property
    def combine_kwargs(self):
        return {"levels": self.levels, "observed": self.observed, "dropna": self.dropna}

    @functools.cached_property
    def chunk_kwargs(self):
        return {"observed": self.observed, "dropna": self.dropna}

    @functools.cached_property
    def aggregate_kwargs(self):
        return {
            "levels": self.levels,
            "sort": self.sort,
            "observed": self.observed,
            "dropna": self.dropna,
        }


def _var_combine(g, levels, sort=False, observed=False, dropna=True):
    return g.groupby(level=levels, sort=sort, observed=observed, dropna=dropna).sum()


class Var(GroupByReduction):
    _parameters = [
        "frame",
        "ddof",
        "numeric_only",
        "split_out",
        "split_every",
        "sort",
        "dropna",
        "observed",
        "shuffle_method",
    ]
    _defaults = {
        "split_out": 1,
        "sort": None,
        "observed": None,
        "dropna": None,
        "split_every": None,
        "shuffle_method": None,
    }
    reduction_aggregate = staticmethod(_var_agg)
    reduction_combine = staticmethod(_var_combine)
    chunk = staticmethod(_var_chunk)

    @functools.cached_property
    def aggregate_kwargs(self):
        return {
            "ddof": self.ddof,
            "numeric_only": self.numeric_only,
            **super().aggregate_kwargs,
        }

    @functools.cached_property
    def chunk_kwargs(self):
        return {"numeric_only": self.numeric_only, **super().chunk_kwargs}


class Std(Var):
    def _lower(self):
        v = Var(*self.operands)
        return MapPartitions(
            v,
            func=np.sqrt,
            meta=v._meta,
            enforce_metadata=True,
            transform_divisions=True,
            clear_divisions=True,
        )


def _mean_chunk(df, *by, observed=None, dropna=None):
    if is_series_like(df):
        df = df.to_frame()

    g = _groupby_raise_unaligned(df, by=by, observed=observed, dropna=dropna)
    x = g.sum(numeric_only=True)
    n = g[x.columns].count().rename(columns=lambda c: c + "-count")
    return concat([x, n], axis=1)


def _mean_combine(g, levels, sort=False, observed=None, dropna=None):
    return g.groupby(level=levels, sort=sort, observed=observed, dropna=dropna).sum()


def _mean_agg(g, levels, sort=False, observed=False, dropna=True):
    result = g.groupby(level=levels, sort=sort, observed=observed, dropna=dropna).sum()
    s = result[result.columns[: len(result.columns) // 2]]
    c = result[result.columns[len(result.columns) // 2 :]]
    c.columns = s.columns
    return s / c


class Mean(GroupByReduction):
    _parameters = SingleAggregation._parameters
    _defaults = SingleAggregation._defaults
    reduction_aggregate = staticmethod(_mean_agg)
    reduction_combine = staticmethod(_mean_combine)
    chunk = staticmethod(_mean_chunk)


def nunique_df_combine(dfs, *args, **kwargs):
    return _nunique_df_combine(concat(dfs), *args, **kwargs)


def nunique_df_aggregate(dfs, levels, name, sort=False):
    df = concat(dfs)
    if df.ndim == 1:
        # split out reduces to a Series
        return df.groupby(level=levels, sort=sort, observed=True).nunique()
    else:
        return df.groupby(level=levels, sort=sort, observed=True)[name].nunique()


class NUnique(SingleAggregation):
    aggregate = staticmethod(nunique_df_aggregate)
    combine = staticmethod(nunique_df_combine)

    @staticmethod
    def chunk(df, *by, **kwargs):
        if df.ndim == 1:
            df = df.to_frame()
            kwargs = dict(name=df.columns[0], levels=_determine_levels(by))
        return _nunique_df_chunk(df, *by, **kwargs)

    @functools.cached_property
    def chunk_kwargs(self) -> dict:  # type: ignore
        kwargs = super().chunk_kwargs
        kwargs["name"] = self._slice
        return kwargs

    @functools.cached_property
    def aggregate_kwargs(self) -> dict:  # type: ignore
        return {"levels": self.levels, "name": self._slice}

    @functools.cached_property
    def combine_kwargs(self):
        return {"levels": self.levels}


class Head(SingleAggregation):
    groupby_chunk = staticmethod(_head_chunk)
    groupby_aggregate = staticmethod(_head_aggregate)

    @classmethod
    def combine(cls, inputs, **kwargs):
        return _concat(inputs)


class Tail(Head):
    groupby_chunk = staticmethod(_tail_chunk)
    groupby_aggregate = staticmethod(_tail_aggregate)


class GroupByApply(Expr, GroupByBase):
    _parameters = [
        "frame",
        "observed",
        "dropna",
        "_slice",
        "group_keys",
        "func",
        "meta",
        "args",
        "kwargs",
        "shuffle_method",
    ]
    _defaults = {
        "observed": None,
        "dropna": None,
        "_slice": None,
        "group_keys": True,
        "shuffle_method": None,
    }

    @functools.cached_property
    def grp_func(self):
        return functools.partial(groupby_slice_apply, func=self.func)

    @functools.cached_property
    def _meta(self):
        if self.operand("meta") is not no_default:
            return make_meta(self.operand("meta"), parent_meta=self.frame._meta)
        return _meta_apply_transform(self, self.grp_func)

    def _divisions(self):
        if self.need_to_shuffle:
            return (None,) * (self.frame.npartitions + 1)
        return self.frame.divisions

    def _shuffle_grp_func(self, shuffled=False):
        return self.grp_func

    @functools.cached_property
    def unique_partition_mapping_columns_from_shuffle(self):
        if not self.need_to_shuffle:
            return self.frame.unique_partition_mapping_columns_from_shuffle
        elif not any(isinstance(b, Expr) for b in self.by):
            return {tuple(self._by_columns)}
        else:
            return set()

    @functools.cached_property
    def need_to_shuffle(self):
        if not any(isinstance(b, Expr) for b in self.by):
            if any(
                set(self._by_columns) >= set(cols)
                for cols in self.frame.unique_partition_mapping_columns_from_shuffle
            ):
                return False

        return any(div is None for div in self.frame.divisions) or not any(
            _contains_index_name(self.frame._meta.index.name, b) for b in self.by
        )

    def _lower(self):
        df = self.frame
        by = self.by

        if self.need_to_shuffle:

            def get_map_columns(df):
                map_columns = {col: str(col) for col in df.columns if col != str(col)}
                unmap_columns = {v: k for k, v in map_columns.items()}
                return map_columns, unmap_columns

            # Map Tuple[str] column names to str before the shuffle

            if any(isinstance(b, Expr) for b in self.by):
                is_series = df.ndim == 1
                if is_series:
                    df = ToFrame(df)
                cols, assign_exprs = [], []
                for i, b in enumerate(self.by):
                    if isinstance(b, Expr):
                        assign_exprs.extend([f"_by_{i}", b])
                        cols.append(f"_by_{i}")
                if len(assign_exprs):
                    df = Assign(df, *assign_exprs)

                map_columns, unmap_columns = get_map_columns(df)
                if map_columns:
                    df = RenameFrame(df, map_columns)

                df = RearrangeByColumn(
                    df,
                    [map_columns.get(c, c) for c in cols],
                    df.npartitions,
                    method=self.shuffle_method,
                )

                if unmap_columns:
                    df = RenameFrame(df, unmap_columns)

                # pandas checks if the group keys are part of the initial
                # DataFrame through reference tracking. This is fine as long
                # as we don't trigger a copy after the Assign above, since the
                # blocks stay separate normally, disk shuffle triggers a copy
                # though, and we shouldn't rely on those internals in pandas,
                # so we can trigger a deep copy here to clear the references
                # since we know more about the query than pandas does.
                by = [
                    (
                        b
                        if not isinstance(b, Expr)
                        else _DeepCopy(
                            RenameSeries(
                                Projection(df, f"_by_{i}"), index=self.by[i].columns[0]
                            )
                        )
                    )
                    for i, b in enumerate(self.by)
                ]
                cols = [col for col in df.columns if col not in cols]
                if is_series:
                    cols = cols[0]
                df = Projection(df, cols)
            else:
                map_columns, unmap_columns = get_map_columns(df)
                if map_columns:
                    df = RenameFrame(df, map_columns)
                df = RearrangeByColumn(
                    df,
                    map_columns.get(self.by[0], self.by[0]),
                    self.npartitions,
                    method=self.shuffle_method,
                )

                if unmap_columns:
                    df = RenameFrame(df, unmap_columns)

            grp_func = self._shuffle_grp_func(True)
        else:
            grp_func = self._shuffle_grp_func(False)

        return GroupByUDFBlockwise(
            df,
            self._slice,
            self.group_keys,
            self.observed,
            self.dropna,
            self.operand("args"),
            self.operand("kwargs"),
            grp_func,
            self.operand("meta"),
            *by,
        )


class GroupByTransform(GroupByApply):
    @functools.cached_property
    def grp_func(self):
        return functools.partial(groupby_slice_transform, func=self.func)


def _fillna(group, *, what, **kwargs):
    return getattr(group, what)(**kwargs)


class GroupByBFill(GroupByTransform):
    func = staticmethod(functools.partial(_fillna, what="bfill"))

    def _simplify_up(self, parent, dependents):
        if isinstance(parent, Projection):
            return groupby_projection(self, parent, dependents)


class GroupByFFill(GroupByBFill):
    func = staticmethod(functools.partial(_fillna, what="ffill"))


class GroupByShift(GroupByApply):
    _defaults = {
        "observed": None,
        "dropna": None,
        "_slice": None,
        "func": None,
        "group_keys": True,
    }

    @functools.cached_property
    def grp_func(self):
        return functools.partial(groupby_slice_shift, shuffled=False)

    def _shuffle_grp_func(self, shuffled=False):
        return functools.partial(groupby_slice_shift, shuffled=shuffled)


class Median(GroupByShift):
    _parameters = GroupByApply._parameters + ["split_every"]
    default = {**GroupByShift._defaults, "split_every": None}

    @functools.cached_property
    def grp_func(self):
        return functools.partial(_median_groupby_aggregate)

    def _shuffle_grp_func(self, shuffled=False):
        return self.grp_func

    def _simplify_up(self, parent, dependents):
        if isinstance(parent, Projection):
            return groupby_projection(self, parent, dependents)

    @functools.cached_property
    def npartitions(self):
        npartitions = self.frame.npartitions
        if self.split_every is not None:
            npartitions = npartitions // self.split_every
        return npartitions


def groupby_get_group(df, *by_key, get_key=None, columns=None):
    if PANDAS_GE_300 and is_scalar(get_key):
        get_key = (get_key,)
    return _groupby_get_group(df, list(by_key), get_key, columns)


class GetGroup(Blockwise, GroupByBase):
    operation = staticmethod(groupby_get_group)
    _parameters = ["frame", "get_key", "columns"]
    _keyword_only = ["get_key", "columns"]

    @property
    def _args(self) -> list:
        return [self.frame] + self.by

    @property
    def _kwargs(self) -> dict:
        cols = self.operand("columns")
        return {
            "get_key": self.get_key,
            "columns": cols if cols is not None else self.frame.columns,
        }


def _median_groupby_aggregate(
    df,
    by=None,
    key=None,
    group_keys=True,  # not used
    dropna=None,
    observed=None,
    numeric_only=False,
    args=None,
    **kwargs,
):
    dropna = {"dropna": dropna} if dropna is not None else {}
    observed = {"observed": observed} if observed is not None else {}

    g = df.groupby(by=by, **observed, **dropna)
    if key is not None:
        g = g[key]
    return g.median(numeric_only=numeric_only)


class GroupByUDFBlockwise(Blockwise, GroupByBase):
    _parameters = [
        "frame",
        "_slice",
        "group_keys",
        "observed",
        "dropna",
        "args",
        "kwargs",
        "dask_func",
        "meta",
    ]
    _defaults = {"observed": None, "dropna": None}
    _keyword_only = [
        "_slice",
        "group_keys",
        "observed",
        "dropna",
        "args",
        "kwargs",
        "dask_func",
    ]

    @property
    def _args(self) -> list:
        return [self.frame] + self.by

    @functools.cached_property
    def _meta(self):
        if self.operand("meta") is not no_default:
            return make_meta(self.operand("meta"), parent_meta=self.frame._meta)
        return _meta_apply_transform(self, self.dask_func)

    def _task(self, name: Key, index: int) -> Task:
        args = [self._blockwise_arg(op, index) for op in self._args]
        kwargs = self._kwargs.copy()
        kwargs.update(
            {
                "_func": self.operation,
                "_meta": self._meta,
            }
        )
        return Task(name, apply_and_enforce, *args, **kwargs)

    @staticmethod
    def operation(
        frame,
        *by,
        _slice,
        group_keys=None,
        observed=None,
        dropna=None,
        args=None,
        kwargs=None,
        dask_func=None,
    ):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        return dask_func(
            frame,
            list(by),
            key=_slice,
            group_keys=group_keys,
            args=args,
            **_as_dict("observed", observed),
            **_as_dict("dropna", dropna),
            **kwargs,
        )


def _contains_index_name(index_name, by):
    if index_name is None:
        return False

    if isinstance(by, Expr):
        return False

    if not is_scalar(by):
        return False

    return index_name == by


def groupby_slice_apply(
    df,
    grouper,
    key,
    func,
    args,
    group_keys=GROUP_KEYS_DEFAULT,
    dropna=None,
    observed=None,
    **kwargs,
):
    return _groupby_slice_apply(
        df,
        grouper,
        key,
        func,
        *args,
        group_keys=group_keys,
        dropna=dropna,
        observed=observed,
        **kwargs,
    )


def groupby_slice_shift(
    df,
    grouper,
    key,
    args,
    shuffled,
    group_keys=GROUP_KEYS_DEFAULT,
    dropna=None,
    observed=None,
    **kwargs,
):
    return _groupby_slice_shift(
        df, grouper, key, shuffled, group_keys, dropna, observed, **kwargs
    )


def groupby_slice_transform(
    df,
    grouper,
    key,
    func,
    args,
    group_keys=GROUP_KEYS_DEFAULT,
    dropna=None,
    observed=None,
    **kwargs,
):
    return _groupby_slice_transform(
        df,
        grouper,
        key,
        func,
        *args,
        group_keys=group_keys,
        dropna=dropna,
        observed=observed,
        **kwargs,
    )


def _meta_apply_transform(obj, grp_func):
    kwargs = obj.operand("kwargs")
    by_meta = obj._by_meta
    by_meta = [x if is_scalar(x) else meta_nonempty(x) for x in by_meta]
    meta_args, meta_kwargs = _extract_meta((obj.operand("args"), kwargs), nonempty=True)
    return make_meta(
        grp_func(
            meta_nonempty(obj.frame._meta),
            by_meta,
            key=obj._slice,
            args=meta_args,
            **_as_dict("observed", obj.observed),
            **_as_dict("dropna", obj.dropna),
            **_as_dict("group_keys", obj.group_keys),
            **meta_kwargs,
        )
    )


def groupby_projection(expr, parent, dependents):
    if isinstance(parent, Projection):
        columns = determine_column_projection(
            expr, parent, dependents, additional_columns=expr._by_columns
        )
        columns = _convert_to_list(columns)
        columns = [col for col in expr.frame.columns if col in columns]
        if columns == expr.frame.columns:
            return
        return type(parent)(
            type(expr)(expr.frame[columns], *expr.operands[1:]),
            *parent.operands[1:],
        )
    return


###
### Groupby Collection API
###


def _clean_by_expr(obj, by):
    if (
        isinstance(by, Series)
        and by.name in obj.columns
        and obj.ndim == 2
        and by._name == obj[by.name]._name
    ):
        return by.name
    elif isinstance(by, Index) and by._name == obj.index._name:
        return by.expr
    elif isinstance(by, (Series, Index)):
        if not are_co_aligned(obj.expr, by.expr):
            raise NotImplementedError(
                "by must be in the DataFrames columns or aligned with the DataFrame."
            )
        if isinstance(by, Index):
            by = by.to_series()
            by.index = obj.index
        return by.expr

    # By is a column name, e.g. str or int
    return by


class GroupByCumulative(Expr, GroupByBase):
    _parameters = ["frame", "dropna", "_slice", "numeric_only"]
    _defaults = {"numeric_only": None, "dropna": None, "_slice": None}
    chunk = None
    aggregate: Callable | None = None
    initial = 0

    @functools.cached_property
    def _meta(self):
        cols = None if self._slice is None else self._slice
        return _apply_chunk(
            self.frame._meta,
            *self._by_meta,
            chunk=self.chunk,
            columns=cols,
            **self.numeric_only,
        )

    def _divisions(self):
        return self.frame.divisions

    @property
    def numeric_only(self):
        no = self.operand("numeric_only")
        return {} if no is None else {"numeric_only": no}

    def _lower(self):
        meta = self._meta
        dropna = {} if self.dropna is None else {"dropna": self.dropna}
        columns = meta.name if is_series_like(meta) else meta.columns

        frame = MapPartitions(
            self.frame,
            _apply_chunk,
            meta,
            True,
            True,
            False,
            True,
            None,
            None,
            None,
            {"chunk": self.chunk, "columns": columns, **dropna, **self.numeric_only},
            len(self.by),
            *self.by,
        )
        cum_raw = frame

        if frame.ndim == 1:
            frame = ToFrame(frame)

        by = self.by.copy()
        for i, b in enumerate(by):
            if not isinstance(b, Expr):
                if b in self.frame.columns:
                    frame = Assign(frame, f"_by_{b}", self.frame[b])
                else:
                    frame = Assign(frame, f"_by_{b}", self.frame.index)

                by[i] = f"_by_{b}"

        columns = 0 if columns is None else columns
        cum_last = MapPartitions(
            frame,
            _apply_chunk,
            no_default,
            True,
            True,
            False,
            True,
            None,
            None,
            None,
            {"chunk": M.last, "columns": columns, **dropna},
            len(by),
            *by,
        )
        return GroupByCumulativeFinalizer(
            frame,
            cum_raw,
            cum_last,
            meta,
            self.aggregate,
            self.initial,
            columns,
            *by,
        )


class GroupByCumulativeFinalizer(Expr, GroupByBase):
    _parameters = [
        "frame",
        "cum_raw",
        "cum_last",
        "meta",
        "aggregate",
        "initial",
        "columns",
    ]

    @functools.cached_property
    def _meta(self):
        return self.meta

    def _divisions(self):
        return self.frame.divisions

    def _layer(self) -> dict:
        dsk = {(self._name, 0): (self.cum_raw._name, 0)}
        name_cum = "cum-last" + self._name

        for i in range(1, self.frame.npartitions):
            # store each cumulative step to graph to reduce computation
            if i == 1:
                dsk[(name_cum, i)] = (self.cum_last._name, i - 1)
            else:
                # aggregate with previous cumulation results
                dsk[(name_cum, i)] = (  # type: ignore
                    _cum_agg_filled,
                    (name_cum, i - 1),
                    (self.cum_last._name, i - 1),
                    self.aggregate,
                    self.initial,
                )
            dsk[(self._name, i)] = (  # type: ignore
                _cum_agg_aligned,
                (self.frame._name, i),
                (name_cum, i),
                self.by,
                self.operand("columns"),
                self.aggregate,
                self.initial,
            )
        return dsk


class GroupByCumsum(GroupByCumulative):
    chunk = M.cumsum
    aggregate = M.add
    initial = 0


class GroupByCumprod(GroupByCumulative):
    chunk = M.cumprod
    aggregate = M.mul
    initial = 1


class GroupByCumcount(GroupByCumulative):
    chunk = M.cumcount
    aggregate = staticmethod(_cumcount_aggregate)
    initial = -1


class GroupBy:
    """Collection container for groupby aggregations

    The purpose of this class is to expose an API similar
    to Pandas' `Groupby` for dask-expr collections.

    See Also
    --------
    SingleAggregation
    """

    def __init__(
        self,
        obj,
        by,
        group_keys=True,
        sort=None,
        observed=None,
        dropna=None,
        slice=None,
    ):
        if isinstance(by, (tuple, list)):
            by = [_clean_by_expr(obj, x) for x in by]
        else:
            by = _clean_by_expr(obj, by)

        by_ = by if isinstance(by, (tuple, list)) else [by]
        if any(isinstance(key, pd.Grouper) for key in by_):
            raise NotImplementedError("pd.Grouper is currently not supported by Dask.")
        self._slice = slice
        # Check if we can project columns
        projection = None
        if (
            np.isscalar(slice)
            or isinstance(slice, (str, list, tuple))
            or (
                (is_index_like(slice) or is_series_like(slice))
                and not is_dask_collection(slice)
            )
        ):
            projection = set(by_).union(
                {slice} if (np.isscalar(slice) or isinstance(slice, str)) else slice
            )
            projection = [c for c in obj.columns if c in projection]

        self.obj = obj[projection] if projection is not None else obj
        self.sort = sort
        self.observed = (
            observed if observed is not None else False if not PANDAS_GE_300 else True
        )
        self.dropna = dropna
        self.group_keys = group_keys
        self.by = (
            [by] if np.isscalar(by) or isinstance(by, (Expr, Callable)) else list(by)
        )
        # surface pandas errors
        self._meta = self.obj._meta.groupby(
            by,
            group_keys=group_keys,
            sort=sort,
            **_as_dict("observed", observed),
            **_as_dict("dropna", dropna),
        )
        if slice is not None:
            if isinstance(slice, tuple):
                slice = list(slice)
            self._meta = self._meta[slice]

    def _numeric_only_kwargs(self, numeric_only):
        kwargs = {"numeric_only": numeric_only}
        return {"chunk_kwargs": kwargs.copy(), "aggregate_kwargs": kwargs.copy()}

    def _single_agg(
        self,
        expr_cls,
        split_every=None,
        split_out=None,
        chunk_kwargs=None,
        aggregate_kwargs=None,
        shuffle_method=None,
    ):
        if split_every is None:
            split_every = 8
        return new_collection(
            expr_cls(
                self.obj.expr,
                self.observed,
                self.dropna,
                chunk_kwargs,
                aggregate_kwargs,
                self._slice,
                split_every,
                split_out,
                self.sort,
                get_specified_shuffle(shuffle_method),
                *self.by,
            )
        )

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError as e:
            raise AttributeError(e) from e

    def __dir__(self):
        return sorted(
            set(
                dir(type(self))
                + list(self.__dict__)
                + list(filter(M.isidentifier, self.obj.columns))
            )
        )

    def compute(self, **kwargs):
        raise NotImplementedError(
            "DataFrameGroupBy does not allow compute method."
            "Please chain it with an aggregation method (like ``.mean()``) or get a "
            "specific group using ``.get_group()`` before calling ``compute()``"
        )

    def __getitem__(self, key):
        if is_scalar(key):
            return SeriesGroupBy(
                self.obj,
                by=self.by,
                group_keys=self.group_keys,
                slice=key,
                sort=self.sort,
                dropna=self.dropna,
                observed=self.observed,
            )
        g = GroupBy(
            self.obj,
            by=self.by,
            slice=key,
            sort=self.sort,
            dropna=self.dropna,
            observed=self.observed,
            group_keys=self.group_keys,
        )
        return g

    @derived_from(
        pd.core.groupby.GroupBy,
        inconsistencies="If the group is not present, Dask will return an empty Series/DataFrame.",
    )
    def get_group(self, key):
        return new_collection(GetGroup(self.obj.expr, key, self._slice, *self.by))

    @derived_from(pd.core.groupby.GroupBy)
    def count(self, **kwargs):
        return self._single_agg(Count, **kwargs)

    @derived_from(pd.core.groupby.GroupBy)
    def sum(self, numeric_only=False, min_count=None, **kwargs):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        result = self._single_agg(Sum, **kwargs, **numeric_kwargs)
        if min_count:
            return result.where(self.count() >= min_count, other=np.nan)
        return result

    @derived_from(pd.core.groupby.GroupBy)
    def prod(self, numeric_only=False, min_count=None, **kwargs):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        result = self._single_agg(Prod, **kwargs, **numeric_kwargs)
        if min_count:
            return result.where(self.count() >= min_count, other=np.nan)
        return result

    def _cum_agg(self, cls, numeric_only=None):
        return new_collection(
            cls(
                self.obj.expr,
                self.dropna,
                self._slice,
                numeric_only,
                *self.by,
            )
        )

    @derived_from(pd.core.groupby.GroupBy)
    def cumsum(self, numeric_only=False):
        return self._cum_agg(GroupByCumsum, numeric_only)

    @derived_from(pd.core.groupby.GroupBy)
    def cumprod(self, numeric_only=False):
        return self._cum_agg(GroupByCumprod, numeric_only)

    @derived_from(pd.core.groupby.GroupBy)
    def cumcount(self):
        return self._cum_agg(GroupByCumcount)

    def _all_numeric(self):
        """Are all columns that we're not grouping on numeric?"""
        numerics = self.obj._meta._get_numeric_data()
        # This computes a groupby but only on the empty meta
        post_group_columns = self._meta.count().columns
        return len(set(post_group_columns) - set(numerics.columns)) == 0

    @derived_from(pd.core.groupby.GroupBy)
    def mean(self, numeric_only=False, split_out=None, **kwargs):
        if not numeric_only and not self._all_numeric():
            raise NotImplementedError(
                "'numeric_only=False' is not implemented in Dask."
            )
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        result = self._single_agg(Mean, split_out=split_out, **kwargs, **numeric_kwargs)
        return self._postprocess_series_squeeze(result)

    def _postprocess_series_squeeze(self, result):
        if (
            isinstance(self.obj, Series)
            or is_scalar(self._slice)
            and self._slice is not None
        ):
            if len(result.columns) < 1:
                raise NotImplementedError(
                    "Cannot call `SeriesGroupBy.var` or `SeriesGroupBy.mean` on the key "
                    "column. Please use `aggregate` if you really need to do this."
                )
            result = result[result.columns[0]]
        return result

    @derived_from(pd.core.groupby.GroupBy)
    def min(self, numeric_only=False, **kwargs):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        return self._single_agg(Min, **kwargs, **numeric_kwargs)

    @derived_from(pd.core.groupby.GroupBy)
    def max(self, numeric_only=False, **kwargs):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        return self._single_agg(Max, **kwargs, **numeric_kwargs)

    @derived_from(pd.core.groupby.GroupBy)
    def first(self, numeric_only=False, sort=None, **kwargs):
        if sort:
            raise NotImplementedError()
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        return self._single_agg(First, **kwargs, **numeric_kwargs)

    @derived_from(pd.DataFrame)
    def cov(
        self,
        ddof=1,
        split_every=None,
        split_out=None,
        numeric_only=False,
        shuffle_method=None,
    ):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        return self._single_agg(
            Cov,
            split_every,
            split_out,
            chunk_kwargs=numeric_kwargs["chunk_kwargs"],
            aggregate_kwargs={"ddof": ddof},
        )

    @derived_from(pd.DataFrame)
    def corr(
        self, split_every=None, split_out=None, numeric_only=False, shuffle_method=None
    ):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        return self._single_agg(
            Corr,
            split_every,
            split_out,
            chunk_kwargs=numeric_kwargs["chunk_kwargs"],
            aggregate_kwargs={"ddof": 1},
        )

    @derived_from(pd.core.groupby.GroupBy)
    def last(self, numeric_only=False, sort=None, **kwargs):
        if sort:
            raise NotImplementedError()
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        return self._single_agg(Last, **kwargs, **numeric_kwargs)

    @derived_from(pd.core.groupby.GroupBy)
    def ffill(self, limit=None, shuffle_method=None):
        return self._transform_like_op(
            GroupByFFill, None, limit=limit, shuffle_method=shuffle_method
        )

    @derived_from(pd.core.groupby.GroupBy)
    def bfill(self, limit=None, shuffle_method=None):
        return self._transform_like_op(
            GroupByBFill, None, limit=limit, shuffle_method=shuffle_method
        )

    @derived_from(pd.core.groupby.GroupBy)
    def size(self, **kwargs):
        return self._single_agg(Size, **kwargs)

    @derived_from(pd.DataFrame)
    def idxmin(
        self,
        split_every=None,
        split_out=None,
        skipna=True,
        numeric_only=False,
        shuffle_method=None,
    ):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        numeric_kwargs["chunk_kwargs"]["skipna"] = skipna
        return self._single_agg(
            IdxMin,
            split_every=split_every,
            split_out=split_out,
            shuffle_method=shuffle_method,
            **numeric_kwargs,
        )

    @derived_from(pd.DataFrame)
    def idxmax(
        self,
        split_every=None,
        split_out=None,
        skipna=True,
        numeric_only=False,
        shuffle_method=None,
    ):
        numeric_kwargs = self._numeric_only_kwargs(numeric_only)
        numeric_kwargs["chunk_kwargs"]["skipna"] = skipna
        return self._single_agg(
            IdxMax,
            split_every=split_every,
            split_out=split_out,
            shuffle_method=shuffle_method,
            **numeric_kwargs,
        )

    @derived_from(pd.core.groupby.SeriesGroupBy)
    def head(self, n=5, split_every=None, split_out=None):
        chunk_kwargs = {"n": n}
        aggregate_kwargs = {
            "n": n,
            "index_levels": len(self.by) if not isinstance(self.by, Expr) else 1,
        }
        return self._single_agg(
            Head,
            split_every=split_every,
            split_out=split_out,
            chunk_kwargs=chunk_kwargs,
            aggregate_kwargs=aggregate_kwargs,
        )

    @derived_from(pd.core.groupby.SeriesGroupBy)
    def tail(self, n=5, split_every=None, split_out=None):
        chunk_kwargs = {"n": n}
        aggregate_kwargs = {
            "n": n,
            "index_levels": len(self.by) if not isinstance(self.by, Expr) else 1,
        }
        return self._single_agg(
            Tail,
            split_every=split_every,
            split_out=split_out,
            chunk_kwargs=chunk_kwargs,
            aggregate_kwargs=aggregate_kwargs,
        )

    @derived_from(pd.core.groupby.GroupBy)
    def var(
        self,
        ddof=1,
        split_every=None,
        split_out=None,
        numeric_only=False,
        shuffle_method=None,
    ):
        if not numeric_only and not self._all_numeric():
            raise NotImplementedError(
                "'numeric_only=False' is not implemented in Dask."
            )
        result = new_collection(
            Var(
                self.obj.expr,
                ddof,
                numeric_only,
                split_out,
                split_every,
                self.sort,
                self.dropna,
                self.observed,
                shuffle_method,
                *self.by,
            )
        )
        return self._postprocess_series_squeeze(result)

    @derived_from(pd.core.groupby.GroupBy)
    def std(
        self,
        ddof=1,
        split_every=None,
        split_out=None,
        numeric_only=False,
        shuffle_method=None,
    ):
        if not numeric_only and not self._all_numeric():
            raise NotImplementedError(
                "'numeric_only=False' is not implemented in Dask."
            )
        result = new_collection(
            Std(
                self.obj.expr,
                ddof,
                numeric_only,
                split_out,
                split_every,
                self.sort,
                self.dropna,
                self.observed,
                shuffle_method,
                *self.by,
            )
        )
        return self._postprocess_series_squeeze(result)

    @_aggregate_docstring(based_on="pd.core.groupby.DataFrameGroupBy.agg")
    def aggregate(
        self, arg=None, split_every=8, split_out=None, shuffle_method=None, **kwargs
    ):
        relabeling, order, columns = None, None, None
        if arg is None:
            if not isinstance(self, SeriesGroupBy):
                relabeling, arg, columns, order = reconstruct_func(arg, **kwargs)
            elif isinstance(self, SeriesGroupBy):
                columns, arg = validate_func_kwargs(kwargs)
                relabeling = True

        if arg == "size":
            return self.size()

        result = new_collection(
            GroupbyAggregation(
                self.obj.expr,
                arg,
                self.observed,
                self.dropna,
                split_every,
                split_out,
                self.sort,
                shuffle_method,
                self._slice,
                *self.by,
            )
        )
        if relabeling and result is not None:
            if order is not None:
                result = result.iloc[:, order]
            result.columns = columns
        return result

    def agg(self, *args, **kwargs):
        return self.aggregate(*args, **kwargs)

    def _warn_if_no_meta(self, meta, method="apply"):
        if meta is no_default:
            msg = f"""`meta` is not specified, inferred from partial data.
Please provide `meta` if the result is unexpected.
  Before: .{method}(func)
  After:  .{method}(func, meta={{'x': 'f8', 'y': 'f8'}}) for dataframe result
  or:     .{method}(func, meta=('x', 'f8'))            for series result
"""
            warnings.warn(msg, stacklevel=3)

    @insert_meta_param_description(pad=12)
    def apply(self, func, *args, meta=no_default, shuffle_method=None, **kwargs):
        """Parallel version of pandas GroupBy.apply

        This mimics the pandas version except for the following:

        1.  If the grouper does not align with the index then this causes a full
            shuffle.  The order of rows within each group may not be preserved.
        2.  Dask's GroupBy.apply is not appropriate for aggregations. For custom
            aggregations, use :class:`dask.dataframe.groupby.Aggregation`.

        .. warning::

           Pandas' groupby-apply can be used to to apply arbitrary functions,
           including aggregations that result in one row per group. Dask's
           groupby-apply will apply ``func`` once on each group, doing a shuffle
           if needed, such that each group is contained in one partition.
           When ``func`` is a reduction, e.g., you'll end up with one row
           per group. To apply a custom aggregation with Dask,
           use :class:`dask.dataframe.groupby.Aggregation`.

        Parameters
        ----------
        func: function
            Function to apply
        args, kwargs : Scalar, Delayed or object
            Arguments and keywords to pass to the function.
        $META

        Returns
        -------
        applied : Series or DataFrame depending on columns keyword
        """
        self._warn_if_no_meta(meta)
        return new_collection(
            GroupByApply(
                self.obj.expr,
                self.observed,
                self.dropna,
                self._slice,
                self.group_keys,
                func,
                meta,
                args,
                kwargs,
                get_specified_shuffle(shuffle_method),
                *self.by,
            )
        )

    def _transform_like_op(
        self, expr_cls, func, meta=no_default, shuffle_method=None, *args, **kwargs
    ):
        return new_collection(
            expr_cls(
                self.obj.expr,
                self.observed,
                self.dropna,
                self._slice,
                self.group_keys,
                func,
                meta,
                args,
                kwargs,
                get_specified_shuffle(shuffle_method),
                *self.by,
            )
        )

    @insert_meta_param_description(pad=12)
    def transform(self, func, meta=no_default, shuffle_method=None, *args, **kwargs):
        """Parallel version of pandas GroupBy.transform

        This mimics the pandas version except for the following:

        1.  If the grouper does not align with the index then this causes a full
            shuffle.  The order of rows within each group may not be preserved.
        2.  Dask's GroupBy.transform is not appropriate for aggregations. For custom
            aggregations, use :class:`dask.dataframe.groupby.Aggregation`.

        .. warning::

           Pandas' groupby-transform can be used to apply arbitrary functions,
           including aggregations that result in one row per group. Dask's
           groupby-transform will apply ``func`` once on each group, doing a shuffle
           if needed, such that each group is contained in one partition.
           When ``func`` is a reduction, e.g., you'll end up with one row
           per group. To apply a custom aggregation with Dask,
           use :class:`dask.dataframe.groupby.Aggregation`.

        Parameters
        ----------
        func: function
            Function to apply
        args, kwargs : Scalar, Delayed or object
            Arguments and keywords to pass to the function.
        $META

        Returns
        -------
        applied : Series or DataFrame depending on columns keyword
        """
        self._warn_if_no_meta(meta, method="transform")
        return self._transform_like_op(
            GroupByTransform, func, meta, shuffle_method, *args, **kwargs
        )

    @insert_meta_param_description(pad=12)
    def shift(self, periods=1, meta=no_default, shuffle_method=None, *args, **kwargs):
        """Parallel version of pandas GroupBy.shift

        This mimics the pandas version except for the following:

        If the grouper does not align with the index then this causes a full
        shuffle.  The order of rows within each group may not be preserved.

        Parameters
        ----------
        periods : Delayed, Scalar or int, default 1
            Number of periods to shift.
        freq : Delayed, Scalar or str, optional
            Frequency string.
        fill_value : Scalar, Delayed or object, optional
            The scalar value to use for newly introduced missing values.
        $META

        Returns
        -------
        shifted : Series or DataFrame shifted within each group.

        Examples
        --------
        >>> import dask
        >>> ddf = dask.datasets.timeseries(freq="1h")
        >>> result = ddf.groupby("name").shift(1, meta={"id": int, "x": float, "y": float})
        """
        if "axis" in kwargs:
            raise TypeError("axis is not supported in shift.")
        self._warn_if_no_meta(meta, method="shift")
        kwargs = {"periods": periods, **kwargs}
        return self._transform_like_op(
            GroupByShift, None, meta, shuffle_method, *args, **kwargs
        )

    @derived_from(pd.core.groupby.GroupBy)
    def median(
        self, split_every=None, split_out=True, shuffle_method=None, numeric_only=False
    ):
        result = new_collection(
            Median(
                self.obj.expr,
                self.observed,
                self.dropna,
                self._slice,
                self.group_keys,
                None,
                no_default,
                (),
                {"numeric_only": numeric_only},
                get_specified_shuffle(shuffle_method),
                split_every,
                *self.by,
            )
        )
        if split_out is not True:
            result = result.repartition(npartitions=split_out)
        return result

    def rolling(self, window, min_periods=None, center=False, win_type=None, axis=0):
        """Provides rolling transformations.

        .. note::

            Since MultiIndexes are not well supported in Dask, this method returns a
            dataframe with the same index as the original data. The groupby column is
            not added as the first level of the index like pandas does.

            This method works differently from other groupby methods. It does a groupby
            on each partition (plus some overlap). This means that the output has the
            same shape and number of partitions as the original.

        Parameters
        ----------
        window : str, offset
           Size of the moving window. This is the number of observations used
           for calculating the statistic. Data must have a ``DatetimeIndex``
        min_periods : int, default None
            Minimum number of observations in window required to have a value
            (otherwise result is NA).
        center : boolean, default False
            Set the labels at the center of the window.
        win_type : string, default None
            Provide a window type. The recognized window types are identical
            to pandas.
        axis : int, default 0

        Returns
        -------
        a Rolling object on which to call a method to compute a statistic

        Examples
        --------
        >>> import dask
        >>> ddf = dask.datasets.timeseries(freq="1h")
        >>> result = ddf.groupby("name").x.rolling('1D').max()
        """
        from dask.dataframe.dask_expr._rolling import Rolling

        return Rolling(
            self.obj,
            window,
            min_periods=min_periods,
            center=center,
            win_type=win_type,
            groupby_kwargs={
                "by": self.by,
                "sort": self.sort,
                "observed": self.observed,
                "dropna": self.dropna,
                "group_keys": self.group_keys,
            },
            groupby_slice=self._slice,
        )


class SeriesGroupBy(GroupBy):
    def __init__(
        self,
        obj,
        by,
        group_keys=True,
        sort=None,
        observed=None,
        dropna=None,
        slice=None,
    ):
        # Raise pandas errors if applicable
        if isinstance(obj, Series):
            if isinstance(by, FrameBase):
                obj._meta.groupby(by._meta, **_as_dict("observed", observed))
            elif isinstance(by, (list, tuple)) and any(
                isinstance(x, FrameBase) for x in by
            ):
                metas = [x._meta if isinstance(x, FrameBase) else x for x in by]
                obj._meta.groupby(metas, **_as_dict("observed", observed))
            elif isinstance(by, list):
                if len(by) == 0:
                    raise ValueError("No group keys passed!")

                non_series_items = [item for item in by if not isinstance(item, Series)]
                obj._meta.groupby(non_series_items, **_as_dict("observed", observed))
            else:
                obj._meta.groupby(by, **_as_dict("observed", observed))

        super().__init__(
            obj,
            by=by,
            group_keys=group_keys,
            slice=slice,
            observed=observed,
            dropna=dropna,
            sort=sort,
        )

    @derived_from(pd.core.groupby.SeriesGroupBy)
    def value_counts(self, **kwargs):
        return self._single_agg(ValueCounts, **kwargs)

    @derived_from(pd.core.groupby.SeriesGroupBy)
    def unique(self, **kwargs):
        return self._single_agg(Unique, **kwargs)

    def idxmin(
        self,
        split_every=None,
        split_out=None,
        skipna=True,
        numeric_only=False,
        **kwargs,
    ):
        # pandas doesn't support numeric_only here, which is odd
        return self._single_agg(
            IdxMin,
            split_every=None,
            split_out=split_out,
            chunk_kwargs=dict(skipna=skipna),
        )

    def idxmax(
        self,
        split_every=None,
        split_out=None,
        skipna=True,
        numeric_only=False,
        **kwargs,
    ):
        # pandas doesn't support numeric_only here, which is odd
        return self._single_agg(
            IdxMax,
            split_every=split_every,
            split_out=split_out,
            chunk_kwargs=dict(skipna=skipna),
        )

    @derived_from(pd.core.groupby.SeriesGroupBy)
    def nunique(self, split_every=None, split_out=True, shuffle_method=None):
        """
        Examples
        --------
        >>> import pandas as pd
        >>> import dask.dataframe as dd
        >>> d = {'col1': [1, 2, 3, 4], 'col2': [5, 6, 7, 8]}
        >>> df = pd.DataFrame(data=d)
        >>> ddf = dd.from_pandas(df, 2)
        >>> ddf.groupby(['col1']).col2.nunique().compute()
        """
        slice = self._slice or self.obj.name
        return new_collection(
            NUnique(
                self.obj.expr,
                self.observed,
                self.dropna,
                None,
                None,
                slice,
                split_every,
                split_out,
                self.sort,
                get_specified_shuffle(shuffle_method),
                *self.by,
            )
        )

    def cov(self, *args, **kwargs):
        raise NotImplementedError("cov is not implemented for SeriesGroupBy objects.")

    def corr(self, *args, **kwargs):
        raise NotImplementedError("cov is not implemented for SeriesGroupBy objects.")

    def _all_numeric(self):
        return True
